# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import io
import os

import disnake
from disnake.ext import commands

from jishaku.codeblocks import codeblock_converter
from jishaku.exception_handling import ReplResponseReactor
from jishaku.features.baseclass import Feature
from jishaku.flags import Flags, DISABLED_SYMBOLS
from jishaku.functools import AsyncSender
from jishaku.paginators import PaginatorInterface, WrappedPaginator, use_file_check
from jishaku.repl import AsyncCodeExecutor, Scope, all_inspections, disassemble, get_var_dict_from_ctx


class PythonFeature(Feature):
    """
    Функция, содержащая команды, связанные с Python
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._scope = Scope()
        self.retain = Flags.RETAIN
        self.last_result = None

    @property
    def scope(self):
        """
        Получает пример для использования в Repl.

        Если удержание включено, это внутренняя сохраненная область,
        В противном случае это всегда новая сфера.
        """

        if self.retain:
            return self._scope
        return Scope()

    @Feature.Command(parent="jsk", name="retain")
    async def jsk_retain(self, ctx: commands.Context, *, toggle: bool = None):
        """
        Поверните переменную удержание для Repl On или Off.

        Не указывать аргументации для текущего статуса.
        """

        if toggle is None:
            if self.retain:
                return await ctx.send("Удержание переменной установлено на ON")

            return await ctx.send("Удержание переменной установлено на OFF")

        if toggle:
            if self.retain:
                return await ctx.send("Удержание переменной уже установлено на ON.")

            self.retain = True
            self._scope = Scope()
            return await ctx.send("Удержание переменной включена. Будущие сеансы Reply сохранят свои действия.")

        if not self.retain:
            return await ctx.send("Удержание переменной уже установлено на OFF.")

        self.retain = False
        return await ctx.send("Удержание переменной выключена. Будущие сессии Reply избавят свои возможности, когда закончите.")

    async def jsk_python_result_handling(self, ctx: commands.Context, result):
        """
        Determines what is done with a result when it comes out of jsk py.
        This allows you to override how this is done without having to rewrite the command itself.
        What you return is what gets stored in the temporary _ variable.
        """

        handle = os.getenv('JISHAKU_PY_RES', 'true')

        if handle not in DISABLED_SYMBOLS:
            if isinstance(result, disnake.Message):
                return await ctx.send(f"<Сообщение <{result.jump_url}>>")

            if isinstance(result, disnake.File):
                return await ctx.send(file=result)

            if isinstance(result, disnake.Embed):
                return await ctx.send(embed=result)

            if isinstance(result, PaginatorInterface):
                return await result.send_to(ctx)

            if not isinstance(result, str):
                # Решить все не-стряхи
                result = repr(result)

        if isinstance(result, str):
            if len(result) <= 2000:
                if result.strip() == '':
                    result = "\u200b"

                return await ctx.send(result.replace(self.bot.http.token, "[token omitted]"))

            if use_file_check(ctx, len(result)):  # Файл "Полный контент" Предел предварительного просмотра
                # Discord's Desktop и веб -клиент теперь поддерживают интерактивный файловый контент
                #  отображать для файлов, кодируемых в UTF-8.
                # Поскольку это избегает проблем с побегом и является более интуитивно понятным, чем страниц для
                #  длинные результаты, теперь он будет расставлен в приоритет по PaginatorInterface, если
                #  Результирующий контент ниже порога размера файлов
                return await ctx.send(file=disnake.File(
                    filename="output.py",
                    fp=io.BytesIO(result.encode('utf-8'))
                ))

            # несоответствие здесь, результаты обернуты в кодовые блоки, когда они слишком большие
            #  Но не так ли, если они нет.Вероятно, не так уж и плохо, но отмечая для последующего обзора
            paginator = WrappedPaginator(prefix='```py', suffix='```', max_size=1985)

            paginator.add_line(result)

            interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
            return await interface.send_to(ctx)

    @Feature.Command(parent="jsk", name="py", aliases=["python"])
    async def jsk_python(self, ctx: commands.Context, *, argument: codeblock_converter):
        """
        Прямая оценка кода Python.
        """

        arg_dict = get_var_dict_from_ctx(ctx, Flags.SCOPE_PREFIX)
        arg_dict["_"] = self.last_result

        scope = self.scope

        try:
            async with ReplResponseReactor(ctx.message):
                with self.submit(ctx):
                    executor = AsyncCodeExecutor(argument.content, scope, arg_dict=arg_dict)
                    async for send, result in AsyncSender(executor):
                        if result is None:
                            continue

                        self.last_result = result

                        send(await self.jsk_python_result_handling(ctx, result))

        finally:
            scope.clear_intersection(arg_dict)

    @Feature.Command(parent="jsk", name="py_inspect", aliases=["pyi", "python_inspect", "pythoninspect"])
    async def jsk_python_inspect(self, ctx: commands.Context, *, argument: codeblock_converter):
        """
        Оценка кода Python с проверкой информации.
        """

        arg_dict = get_var_dict_from_ctx(ctx, Flags.SCOPE_PREFIX)
        arg_dict["_"] = self.last_result

        scope = self.scope

        try:
            async with ReplResponseReactor(ctx.message):
                with self.submit(ctx):
                    executor = AsyncCodeExecutor(argument.content, scope, arg_dict=arg_dict)
                    async for send, result in AsyncSender(executor):
                        self.last_result = result

                        header = repr(result).replace("``", "`\u200b`").replace(self.bot.http.token, "[token omitted]")

                        if len(header) > 485:
                            header = header[0:482] + "..."

                        lines = [f"=== {header} ===", ""]

                        for name, res in all_inspections(result):
                            lines.append(f"{name:16.16} :: {res}")

                        text = "\n".join(lines)

                        if use_file_check(ctx, len(text)):  # Файл "Полный контент" Предел предварительного просмотра
                            send(await ctx.send(file=disnake.File(
                                filename="inspection.prolog",
                                fp=io.BytesIO(text.encode('utf-8'))
                            )))
                        else:
                            paginator = WrappedPaginator(prefix="```prolog", max_size=1985)

                            paginator.add_line(text)

                            interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
                            send(await interface.send_to(ctx))
        finally:
            scope.clear_intersection(arg_dict)

    @Feature.Command(parent="jsk", name="dis", aliases=["disassemble"])
    async def jsk_disassemble(self, ctx: commands.Context, *, argument: codeblock_converter):
        """
        Разберите код Python в Bytecode.
        """

        arg_dict = get_var_dict_from_ctx(ctx, Flags.SCOPE_PREFIX)

        async with ReplResponseReactor(ctx.message):
            text = "\n".join(disassemble(argument.content, arg_dict=arg_dict))

            if use_file_check(ctx, len(text)):  # Файл "Полный контент" Предел предварительного просмотра
                await ctx.send(file=disnake.File(
                    filename="dis.py",
                    fp=io.BytesIO(text.encode('utf-8'))
                ))
            else:
                paginator = WrappedPaginator(prefix='```py', max_size=1985)

                paginator.add_line(text)

                interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
                await interface.send_to(ctx)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import dataclasses
import inspect
import os
import typing

ENABLED_SYMBOLS = ("true", "t", "yes", "y", "on", "1")
DISABLED_SYMBOLS = ("false", "f", "no", "n", "off", "0")


@dataclasses.dataclass
class Flag:
    """
    DataClass, который представляет состояние флага Джишаку.Только для внутреннего использования.
    """

    name: str
    flag_type: type
    default: typing.Callable = None
    override: typing.Any = None

    def resolve(self, flags):
        """
        Установите этот флаг.Только для внутреннего использования.
        """

        # Переопределение ручного, игнорируйте окружающую среду в этом случае
        if self.override is not None:
            return self.override

        # Решить из окружающей среды
        env_value = os.getenv(f"JISHAKU_{self.name}", "").strip()

        if env_value:
            if self.flag_type is bool:
                if env_value.lower() in ENABLED_SYMBOLS:
                    return True
                if env_value.lower() in DISABLED_SYMBOLS:
                    return False
            else:
                return self.flag_type(env_value)

        # Запасной, если не разрешается резольвации от окружающей среды
        if self.default is not None:
            if inspect.isfunction(self.default):
                return self.default(flags)

            return self.default

        return self.flag_type()


class FlagMeta(type):
    """
    Metaclass для флагов.
    Это обрабатывает справедливую оценку флагов, позволяя переопределить их во время исполнения.
    """

    def __new__(cls, name, base, attrs):
        attrs['flag_map'] = {}

        for flag_name, flag_type in attrs['__annotations__'].items():
            attrs['flag_map'][flag_name] = Flag(flag_name, flag_type, attrs.pop(flag_name, None))

        return super(FlagMeta, cls).__new__(cls, name, base, attrs)

    def __getattr__(cls, name: str):
        if hasattr(cls, 'flag_map') and name in cls.flag_map:
            return cls.flag_map[name].resolve(cls)

        return super().__getattribute__(name)

    def __setattr__(cls, name: str, value):
        if name in cls.flag_map:
            flag = cls.flag_map[name]

            if not isinstance(value, flag.flag_type):
                raise ValueError(f"Попытка установить флаг {name} наведите {type(value).__name__} (должен быть {flag.flag_type.__name__})")

            flag.override = value
        else:
            super().__setattr__(name, value)


class Flags(metaclass=FlagMeta):
    """
    Флаги для Джишаку.

    Вы можете переопределить их либо через окружающую среду, например:
        export JISHAKU_HIDE=1
    Или вы можете переопределить их программно:
        jishaku.Flags.HIDE = True
    """

    # Флаг, чтобы указать, что группа командной команды Jishaku должна быть спрятана
    HIDE: bool

    # Флаг, чтобы указать, что режим удержания для Repl должен быть включен по умолчанию
    RETAIN: bool

    # Флаг, чтобы указать, что мета -переменные в Repl не следует префикс с подчеркиванием.
    NO_UNDERSCORE: bool

    # Префикс применения, то есть префикс, который появляется перед встроенными переменными Джишаку в сессиях.
    # Рекомендуется установить это программно.
    SCOPE_PREFIX: str = lambda flags: '' if flags.NO_UNDERSCORE else '_'

    # Флаг, чтобы указать, всегда ли использовать Paginators, а не предварительный просмотр файлов Discord
    FORCE_PAGINATOR: bool

    # Флаг, чтобы указать на многословную ошибку трассировки следует отправлять на канал вызова, а не через прямое сообщение.
    NO_DM_TRACEBACK: bool

    # Флаг, чтобы указать использование Braille J в команде выключения
    USE_BRAILLE_J: bool



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import typing

import disnake
from disnake.ext import commands

from jishaku.features.baseclass import Feature


class GuildFeature(Feature):
    """
    Функция, содержащая команды, связанные с гильдией
    """

    @staticmethod
    def apply_overwrites(permissions: dict, allow: int, deny: int, name: str):
        """
        Применяется перезаписывается к словарю разрешений (см. Permtrace),
на основе допустимой и отрицательной маски.
        """

        allow: disnake.Permissions = disnake.Permissions(allow)
        deny: disnake.Permissions = disnake.Permissions(deny)

        # Отрицает первым ..
        for key, value in dict(deny).items():
            # Убедитесь, что в этом отрицается, и это еще не опровергнуто
            # (Мы хотим показать причину отрицания самого низкого уровня)
            if value and permissions[key][0]:
                permissions[key] = (False, f"Это перезапись {name} канала")

        # Затем позволяет
        for key, value in dict(allow).items():
            # Убедитесь, что это разрешено, и это еще не разрешено
            # (Мы хотим показать причину пособия на самом низком уровне)
            if value and not permissions[key][0]:
                permissions[key] = (True, f"Это перезапись {name} канала")

    @staticmethod
    def chunks(array: list, chunk_size: int):
        """
        Кусочки списка в куски данного размера.
        Вероятно, если честно, должно быть в Utils.
        """
        for i in range(0, len(array), chunk_size):
            yield array[i:i + chunk_size]

    @Feature.Command(parent="jsk", name="permtrace")
    async def jsk_permtrace(
        self, ctx: commands.Context,
        channel: typing.Union[disnake.TextChannel, disnake.VoiceChannel],
        *targets: typing.Union[disnake.Member, disnake.Role]
    ):
        """
        Рассчитывает источник предоставленных или отклоненных разрешений.

        Это принимает канал и либо участник, либо список ролей.
        Он рассчитывает разрешения так же, как и Discord, отслеживая источник.
        """

        member_ids = {target.id: target for target in targets if isinstance(target, disnake.Member)}
        roles = []

        for target in targets:
            if isinstance(target, disnake.Member):
                roles.extend(list(target.roles))
            else:
                roles.append(target)

        # Удалить дубликаты
        roles = list(set(roles))

        # Словарь для хранения текущего состояния разрешения и разума
        # Магазины <имя перми>: (<perm разрешено>, <sonse>)
        permissions: typing.Dict[str, typing.Tuple[bool, str]] = {}

        if member_ids and channel.guild.owner_id in member_ids:
            # Владелец, имеет все пважицы
            for key in dict(disnake.Permissions.all()).keys():
                permissions[key] = (True, f"{channel.guild.owner.mention} Владеет сервером")
        else:
            # В противном случае, либо не участник, либо не владелец гильдии, рассчитывайте вручную вручную
            is_administrator = False

            # Сначала обращаться
            for key, value in dict(channel.guild.default_role.permissions).items():
                permissions[key] = (value, "Это разрешение для всего сервера.")

            for role in roles:
                for key, value in dict(role.permissions).items():
                    # Роли могут только разрешать разрешения
                    # Отказ в разрешении ничего не делает, если более низкая роль позволяет это
                    if value and not permissions[key][0]:
                        permissions[key] = (value, f"это разрешение для всего сервера {role.name}")

                # Затем управляемость администратора
                if role.permissions.administrator:
                    is_administrator = True

                    for key in dict(disnake.Permissions.all()).keys():
                        if not permissions[key][0]:
                            permissions[key] = (True, f"это разрешение предоставляется администратором на уровне всего сервера {role.name}")

            # Если администратор был предоставлен, нет причин даже выполнять разрешения на каналы
            if not is_administrator:
                # Теперь разрешения на уровне канала

                # Особый случай для @EveryOne
                try:
                    maybe_everyone = channel._overwrites[0]
                    if maybe_everyone.id == channel.guild.default_role.id:
                        self.apply_overwrites(permissions, allow=maybe_everyone.allow, deny=maybe_everyone.deny, name="@everyone")
                        remaining_overwrites = channel._overwrites[1:]
                    else:
                        remaining_overwrites = channel._overwrites
                except IndexError:
                    remaining_overwrites = channel._overwrites

                role_lookup = {r.id: r for r in roles}

                # Отрицания применяются ранее, всегда
                # Ручка отрицает
                for overwrite in remaining_overwrites:
                    if overwrite.type == 'role' and overwrite.id in role_lookup:
                        self.apply_overwrites(permissions, allow=0, deny=overwrite.deny, name=role_lookup[overwrite.id].name)

                # Ручка разрешает
                for overwrite in remaining_overwrites:
                    if overwrite.type == 'role' and overwrite.id in role_lookup:
                        self.apply_overwrites(permissions, allow=overwrite.allow, deny=0, name=role_lookup[overwrite.id].name)

                if member_ids:
                    # Обрабатывать специфичные для члена перезаписывания
                    for overwrite in remaining_overwrites:
                        if overwrite.type == 'member' and overwrite.id in member_ids:
                            self.apply_overwrites(permissions, allow=overwrite.allow, deny=overwrite.deny, name=f"{member_ids[overwrite.id].mention}")
                            break

        # Конструкция встроена
        description = f"Это расчет разрешений для следующих целевых объектов в {channel.mention}:\n"
        description += "\n".join(f"- {target.mention}" for target in targets)

        description += (
            "\nПожалуйста, обратите внимание, что указанные причины являются ** наиболее фундаментальной ** причиной, по которой разрешение является таким, какое оно есть. "
            "Могут быть и другие причины, по которым эти разрешения сохраняются, даже если вы измените отображаемые данные."
        )

        embed = disnake.Embed(color=0x00FF00, description=description)

        allows = []
        denies = []

        for key, value in permissions.items():
            if value[0]:
                allows.append(f"\N{WHITE HEAVY CHECK MARK} {key} (потому что {value[1]})")
            else:
                denies.append(f"\N{CROSS MARK} {key} (потому что {value[1]})")

        for chunk in self.chunks(sorted(allows) + sorted(denies), 8):
            embed.add_field(name="...", value="\n".join(chunk), inline=False)

        await ctx.send(embed=embed)



# -*- coding: utf-8 -*-

"""
jishaku.hljs test
~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import asyncio
import inspect

import pytest

from jishaku.functools import executor_function


def sig(*args, **kwargs):
    """
    Вернуть подпись, чтобы облегчить параметризация этого теста
    """
    return args, kwargs


@pytest.mark.parametrize(
    ("args", "kwargs", "expected_return"),
    [
        (*sig(1, 2, c=3), (1, 2, 3)),
        (*sig(3, c=4), (3, None, 4)),
        (*sig(a=5, b=6, c=7), (5, 6, 7))
    ]
)
def test_magic_executor(args, kwargs, expected_return):
    loop = asyncio.get_event_loop()

    def non_executor(a, b=None, *, c) -> tuple:
        return a, b, c

    exact_executor = executor_function(non_executor)

    @executor_function
    def redefined_executor(a, b=None, *, c) -> tuple:
        return a, b, c

    assert inspect.signature(non_executor) == inspect.signature(exact_executor)
    assert inspect.signature(non_executor) == inspect.signature(redefined_executor)

    assert non_executor(*args, **kwargs) == expected_return
    assert loop.run_until_complete(exact_executor(*args, **kwargs)) == expected_return
    assert loop.run_until_complete(redefined_executor(*args, **kwargs)) == expected_return



# -*- coding: utf-8 -*-

"""
jishaku.inspections test
~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import collections  # for __iadd__ test

import disnake
import pytest
from utils import run_async

from jishaku.repl.inspections import all_inspections


@pytest.mark.parametrize(
    "target",
    [
        4,
        disnake.Client,  # Подклассы типа крышки
        tuple,  # Покрыть усечение многократного класса
        [False, 1, "2", 3.0],  # Покрыть типы контента
        collections.Counter,  # Накрыть операторов на место
        run_async  # Покрыть текущие проверки режиссера
    ]
)
def test_object_inspection(target):
    for _, _ in all_inspections(target):
        pass



# -*- coding: utf-8 -*-

"""
jishaku.shell test
~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import asyncio
import sys

import pytest
from utils import run_async

from jishaku.shell import ShellReader


@run_async
async def test_reader_basic():
    return_data = []

    with ShellReader("echo hi") as reader:
        async for result in reader:
            return_data.append(result)

    assert len(return_data) == 1
    assert return_data[0] == "hi"

    with pytest.raises(asyncio.TimeoutError):
        with ShellReader("sleep 2", timeout=1) as reader:
            async for result in reader:
                pass


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Тесты с синтаксисом SH только Linux."
)
@run_async
async def test_linux():
    return_data = []

    with ShellReader(">&2 echo oops") as reader:
        async for result in reader:
            return_data.append(result)

    assert len(return_data) == 1
    assert return_data[0] == "[stderr] oops"

    return_data = []

    with ShellReader("echo one && echo two") as reader:
        async for result in reader:
            return_data.append(result)

    assert len(return_data) == 2
    assert return_data[0] == "one"
    assert return_data[1] == "two"


@pytest.mark.skipif(
    sys.platform != "win32",
    reason="Тесты с синтаксисом CMD только для Windows"
)
@run_async
async def test_windows():
    return_data = []

    with ShellReader("cmd /c \"echo one && echo two && echo three 1>&2\"") as reader:
        async for result in reader:
            return_data.append(result)

    assert len(return_data) == 3
    return_data = [x.strip() for x in return_data]

    assert "one" in return_data
    assert "two" in return_data
    assert "[stderr] three" in return_data



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import ast
import dis

import import_expression

from jishaku.repl.scope import Scope

CORO_CODE = """
import asyncio

import disnake
from disnake.ext import commands
from importlib import import_module as {0}

import jishaku

async def _repl_coroutine({{0}}):
    pass
""".format(import_expression.constants.IMPORTER)


def wrap_code(code: str, args: str = '') -> ast.Module:
    """
    Обертывает код для разобравания.

    Это похоже на функцию с эквивалентом jishaku.repl.compilation,
    но из -за различной структуры, необходимой для чистых разборных,
    Он реализован отдельно здесь.
    """

    user_code = import_expression.parse(code, mode='exec')
    mod = import_expression.parse(CORO_CODE.format(args), mode='exec')

    definition = mod.body[-1]  # async def ...:
    assert isinstance(definition, ast.AsyncFunctionDef)

    # Исправить код пользователя непосредственно в функцию
    definition.body = user_code.body

    ast.fix_missing_locations(mod)

    # Мы не используем здесь трансформатор ключевых слов, поскольку он может привести к вводящей в заблуждение разборке.

    is_asyncgen = any(isinstance(node, ast.Yield) for node in ast.walk(definition))
    last_expr = definition.body[-1]

    # Если последняя часть не является выражением, игнорируйте это
    if not isinstance(last_expr, ast.Expr):
        return mod

    # Если это не генератор, а последнее выражение не является возвратом
    if not is_asyncgen and not isinstance(last_expr.value, ast.Return):
        # скопировать значение выражения в возврат
        return_stmt = ast.Return(last_expr.value)
        ast.copy_location(return_stmt, last_expr)

        # Поместите возвращение, где было первоначальное выражение
        definition.body[-1] = return_stmt

    return mod


def disassemble(code: str, scope: Scope = None, arg_dict: dict = None):
    """
    Разборки асинхронный код в инструкции по байт-коду в стиле dis.dis.
    """

    # Похоже на AsyncCodeExecutor.__init__
    arg_names = list(arg_dict.keys()) if arg_dict else []

    scope = scope or Scope()

    wrapped = wrap_code(code, args=', '.join(arg_names))
    exec(compile(wrapped, '<repl>', 'exec'), scope.globals, scope.locals)

    func_def = scope.locals.get('_repl_coroutine') or scope.globals['_repl_coroutine']

    co = func_def.__code__

    for instruction in dis._get_instructions_bytes(
        co.co_code, co.co_varnames, co.co_names, co.co_consts,
        co.co_cellvars + co.co_freevars, dict(dis.findlinestarts(co)),
        line_offset=0
    ):
        if instruction.starts_line is not None and instruction.offset > 0:
            yield ''

        yield instruction._disassemble(
            4, False, 4
        )



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

from disnake.ext import commands

from jishaku.features.filesystem import FilesystemFeature
from jishaku.features.guild import GuildFeature
from jishaku.features.invocation import InvocationFeature
from jishaku.features.management import ManagementFeature
from jishaku.features.python import PythonFeature
from jishaku.features.root_command import RootCommand
from jishaku.features.shell import ShellFeature
from jishaku.features.voice import VoiceFeature

__all__ = (
    "Jishaku",
    "STANDARD_FEATURES",
    "OPTIONAL_FEATURES",
    "setup",
)

STANDARD_FEATURES = (VoiceFeature, GuildFeature, FilesystemFeature, InvocationFeature, ShellFeature, PythonFeature, ManagementFeature, RootCommand)

OPTIONAL_FEATURES = []

try:
    from jishaku.features.youtube import YouTubeFeature
except ImportError:
    pass
else:
    OPTIONAL_FEATURES.insert(0, YouTubeFeature)


class Jishaku(*OPTIONAL_FEATURES, *STANDARD_FEATURES):
    """
    Подкласс Frontend, который смешивается, чтобы сформировать последний Cog Jishaku.
    """


def setup(bot: commands.Bot):
    """
    Функция настройки, определяющая расширения Jishaku.cog и Jishaku.
    """

    bot.add_cog(Jishaku(bot=bot))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import asyncio

import disnake
from disnake import ui
from disnake.ext import commands

from jishaku.shim.paginator_base import EMOJI_DEFAULT


class PaginatorInterface(ui.View):
    """
    Интерфейс на основе сообщений и реакции для пагинаторов.

    Это позволяет пользователям интерактивно ориентироваться в страницах Paginator и поддерживает живые выводы.

    Пример того, как использовать это со стандартным язычником:

    .. code:: python3

        from disnake.ext import commands

        from jishaku.paginators import PaginatorInterface

        # Где -то в команде ...
            # Паниторы должны иметь уменьшенный MAX_SIZE, чтобы приспособить дополнительный текст, добавленный интерфейсом.
            paginator = commands.Paginator(max_size=1900)

            # Заполнить пагинатора некоторой информацией
            for line in range(100):
                paginator.add_line(f"Line {line + 1}")

            # Создать и отправить интерфейс.
            # Поле «владельца» определяет, кто может взаимодействовать с этим интерфейсом.Если это не так, кто -нибудь может использовать его.
            interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
            await interface.send_to(ctx)

            # send_to создает задачу и возвращает поток управления.
            # Он поднимется, если интерфейс не может быть создан, например, если в канале нет разрешения на реакцию.
            # После того, как интерфейс был отправлен, дополнения линии должны быть сделаны асинхронно, поэтому интерфейс можно обновить.
            await interface.add_line("My, the Earth sure is full of things!")

            # Вы также можете проверить, закрыт ли он, используя свойство «закрыто».
            if not interface.closed:
                await interface.add_line("I'm still here!")
    """

    def __init__(self, bot: commands.Bot, paginator: commands.Paginator, **kwargs):
        if not isinstance(paginator, commands.Paginator):
            raise TypeError('Пагинатор должен быть командами.')

        self._display_page = 0

        self.bot = bot

        self.message = None
        self.paginator = paginator

        self.owner = kwargs.pop('owner', None)
        self.emojis = kwargs.pop('emoji', EMOJI_DEFAULT)
        self.timeout = kwargs.pop('timeout', 7200)
        self.delete_message = kwargs.pop('delete_message', False)

        self.sent_page_reactions = False

        self.task: asyncio.Task = None
        self.send_lock: asyncio.Event = asyncio.Event()

        self.close_exception: Exception = None

        if self.page_size > self.max_page_size:
            raise ValueError(
                f'Парень с патентором имеет слишком большой размер страницы для этого интерфейса. '
                f'({self.page_size} > {self.max_page_size})'
            )

        super().__init__(timeout=self.timeout)

    @property
    def pages(self):
        """
        Возвращает страницы Paginator без преждевременного закрытия активной страницы.
        """
        # Защищенный доступ должен быть разрешен здесь, чтобы не закрывать страницы пагинатора

        paginator_pages = list(self.paginator._pages)
        if len(self.paginator._current_page) > 1:
            paginator_pages.append('\n'.join(self.paginator._current_page) + '\n' + (self.paginator.suffix or ''))

        return paginator_pages

    @property
    def page_count(self):
        """
        Возвращает количество страниц внутреннего пагинатора.
        """

        return len(self.pages)

    @property
    def display_page(self):
        """
        Возвращает текущую страницу. Включен интерфейс Paginator.
        """

        self._display_page = max(0, min(self.page_count - 1, self._display_page))
        return self._display_page

    @display_page.setter
    def display_page(self, value):
        """
        Устанавливает текущую страницу, включенная на пенсионере.Автоматически выдвигает значения входящих.
        """

        self._display_page = max(0, min(self.page_count - 1, value))

    max_page_size = 2000

    @property
    def page_size(self) -> int:
        """
        Свойство, которое возвращает, насколько большая страница, рассчитанная на основе свойств.

        Если это превышает `max_page_size`, исключение поднимается при экземпляре.
        """
        page_count = self.page_count
        return self.paginator.max_size + len(f'\nPage {page_count}/{page_count}')

    @property
    def send_kwargs(self) -> dict:
        """
        Свойство, которое возвращает Kwargs, отправленные для отправки/редактирования при обновлении страницы.

        Поскольку это должно быть совместимо с обоими `disnake.TextChannel.send` и `disnake.Message.edit`,
        это должен быть дикт, содержащий 'content', 'embed' или оба.
        """

        content = self.pages[self.display_page]
        return {'content': content, 'view': self}

    def update_view(self):
        """
        Обновления кнопок просмотра, соответствующие современному состоянию интерфейса.
        Это используется внутри.
        """

        self.button_start.label = f"1 \u200b {self.emojis.start}"
        self.button_previous.label = self.emojis.back
        self.button_current.label = str(self.display_page + 1)
        self.button_next.label = self.emojis.forward
        self.button_last.label = f"{self.emojis.end} \u200b {self.page_count}"
        self.button_close.label = f"{self.emojis.close} \u200b Закрыть пагинатор"

    async def add_line(self, *args, **kwargs):
        """
        Прокси -функция, которая позволяет этому PaginatorInterface оставаться заблокированной на последнюю страницу
        Если это уже на нем.
        """

        display_page = self.display_page
        page_count = self.page_count

        self.paginator.add_line(*args, **kwargs)

        new_page_count = self.page_count

        if display_page + 1 == page_count:
            # Чтобы сохранить позицию фиксированной в конце, обновить позицию на новую последнюю страницу и обновить сообщение.
            self._display_page = new_page_count

        # Безоговорочно установленные отправить блокировку, чтобы попробовать и гарантировать обновления страницы на нефокусированных страницах
        self.send_lock.set()

    async def send_to(self, destination: disnake.abc.Messageable):
        """
        Отправляет сообщение в заданный пункт назначения с этим интерфейсом.

        Это автоматически создает задачу ответа для вас.
        """

        self.message = await destination.send(**self.send_kwargs)

        self.send_lock.set()

        if self.task:
            self.task.cancel()

        self.task = self.bot.loop.create_task(self.wait_loop())

        return self

    @property
    def closed(self):
        """
        Этот интерфейс закрыт?
        """

        if not self.task:
            return False
        return self.task.done()

    async def send_lock_delayed(self):
        """
        Корука, которая возвращается через 1 секунду после выпуска блокировки отправки
        Это помогает уменьшить выброс спама, который быстро достигает ограничений.
        """

        gathered = await self.send_lock.wait()
        self.send_lock.clear()
        await asyncio.sleep(1)
        return gathered

    async def wait_loop(self):
        """
        Ждет в цикле для обновлений в интерфейсе.Это не следует называть вручную - это обрабатывается `send_to`.
        """

        try:
            while not self.bot.is_closed():
                await asyncio.wait_for(self.send_lock_delayed(), timeout=self.timeout)

                self.update_view()

                try:
                    await self.message.edit(**self.send_kwargs)
                except disnake.NotFound:
                    # произошло что -то ужасное
                    return

        except (asyncio.CancelledError, asyncio.TimeoutError) as exception:
            self.close_exception = exception

            if self.bot.is_closed():
                # Ничего не могу сделать с сообщениями, так что просто закройте, чтобы избежать шумной ошибки
                return

            # Если сообщение уже было удалено, эта часть не нужна
            if not self.message:
                return

            if self.delete_message:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    async def interaction_check(self, interaction: disnake.Interaction):
        """Проверьте, что определяет, должно ли это взаимодействие"""
        return not self.owner or interaction.user.id == self.owner.id

    @ui.button(label="1 \u200b \N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}", style=disnake.ButtonStyle.secondary)
    async def button_start(self, button: ui.Button, interaction: disnake.Interaction):
        """Кнопка для отправки интерфейса на первую страницу"""

        self._display_page = 0
        self.update_view()
        await interaction.response.edit_message(**self.send_kwargs)

    @ui.button(label="\N{BLACK LEFT-POINTING TRIANGLE}", style=disnake.ButtonStyle.secondary)
    async def button_previous(self, button: ui.Button, interaction: disnake.Interaction):
        """Кнопка для отправки интерфейса на предыдущую страницу"""

        self._display_page -= 1
        self.update_view()
        await interaction.response.edit_message(**self.send_kwargs)

    @ui.button(label="1", style=disnake.ButtonStyle.primary)
    async def button_current(self, button: ui.Button, interaction: disnake.Interaction):
        """Кнопка, чтобы обновить интерфейс"""

        self.update_view()
        await interaction.response.edit_message(**self.send_kwargs)

    @ui.button(label="\N{BLACK RIGHT-POINTING TRIANGLE}", style=disnake.ButtonStyle.secondary)
    async def button_next(self, button: ui.Button, interaction: disnake.Interaction):
        """Кнопка для отправки интерфейса на следующую страницу"""

        self._display_page += 1
        self.update_view()
        await interaction.response.edit_message(**self.send_kwargs)

    @ui.button(label="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR} \u200b 1", style=disnake.ButtonStyle.secondary)
    async def button_last(self, button: ui.Button, interaction: disnake.Interaction):
        """Кнопка для отправки интерфейса на последнюю страницу"""

        self._display_page = self.page_count - 1
        self.update_view()
        await interaction.response.edit_message(**self.send_kwargs)

    @ui.button(label="\N{BLACK SQUARE FOR STOP} \u200b Закрыть пагинатор", style=disnake.ButtonStyle.danger)
    async def button_close(self, button: ui.Button, interaction: disnake.Interaction):
        """Кнопка для закрытия интерфейса"""

        message = self.message
        self.message = None
        self.task.cancel()
        self.stop()
        await message.delete()


class PaginatorEmbedInterface(PaginatorInterface):
    """
    Подкласс :class:`PaginatorInterface` Это заключает содержание во вставку.
    """

    def __init__(self, *args, **kwargs):
        self._embed = kwargs.pop('embed', None) or disnake.Embed()
        super().__init__(*args, **kwargs)

    @property
    def send_kwargs(self) -> dict:
        self._embed.description = self.pages[self.display_page]
        return {'embed': self._embed, 'view': self}

    max_page_size = 2048

    @property
    def page_size(self) -> int:
        return self.paginator.max_size



# -*- coding: utf-8 -*-

"""
jishaku subclassing functionality test
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import pytest
import utils
from disnake.ext import commands


@pytest.fixture(
    scope='module',
    params=[
        # Subclass 1 (Feature)
        ("tests.subclassed_module_1", "Magnet1", "overridden with a third party feature", commands.Bot, {}),
        ("tests.subclassed_module_1", "Magnet1", "overridden with a third party feature", commands.Bot, {"shard_id": 0, "shard_count": 2}),
        ("tests.subclassed_module_1", "Magnet1", "overridden with a third party feature", commands.AutoShardedBot, {}),
        # Subclass 2 (direct)
        ("tests.subclassed_module_2", "Magnet2", "overridden directly", commands.Bot, {}),
        ("tests.subclassed_module_2", "Magnet2", "overridden directly", commands.Bot, {"shard_id": 0, "shard_count": 2}),
        ("tests.subclassed_module_2", "Magnet2", "overridden directly", commands.AutoShardedBot, {}),
        # Test that the original still works after the load test
        ("jishaku", "Jishaku", "Module was loaded", commands.Bot, {}),
        ("jishaku", "Jishaku", "Module was loaded", commands.Bot, {"shard_id": 0, "shard_count": 2}),
        ("jishaku", "Jishaku", "Module was loaded", commands.AutoShardedBot, {}),
    ],
    ids=[
        "Feature-based subclass (Bot, unsharded)",
        "Feature-based subclass (Bot, sharded)",
        "Feature-based subclass (AutoShardedBot)",
        "direct subclass (Bot, unsharded)",
        "direct subclass (Bot, sharded)",
        "direct subclass (AutoShardedBot)",
        "native (Bot, unsharded)",
        "native (Bot, sharded)",
        "native (AutoShardedBot)"
    ]
)
def bot(request):
    b = request.param[3]('?', **request.param[4])
    b.load_extension(request.param[0])

    b.test_cog = request.param[1]
    b.test_predicate = request.param[2]

    yield b

    b.unload_extension(request.param[0])
    b.loop.run_until_complete(b.close())


@utils.run_async
async def test_commands(bot):
    cog = bot.get_cog(bot.test_cog)

    assert cog is not None

    # тест 'jsk'
    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk').callback(cog, ctx)

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert bot.test_predicate in text



# -*- coding: utf-8 -*-

"""
jishaku converter test
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import asyncio
import inspect
from io import BytesIO

import disnake
import pytest
import utils
from disnake.ext import commands

from jishaku.paginators import FilePaginator, PaginatorEmbedInterface, PaginatorInterface, WrappedPaginator


def test_file_paginator():

    base_text = inspect.cleandoc("""
    #!/usr/bin/env python
    # -*- coding: utf-8 -*-
    pass  # \u3088\u308d\u3057\u304f
    """)

    # Проверьте стандартное кодирование
    pages = FilePaginator(BytesIO(base_text.encode("utf-8"))).pages

    assert len(pages) == 1
    assert pages[0] == f"```python\n{base_text}\n```"

    # Тестовые линии
    pages = FilePaginator(BytesIO(base_text.encode("utf-8")), line_span=(2, 2)).pages

    assert len(pages) == 1
    assert pages[0] == "```python\n# -*- coding: utf-8 -*-\n```"

    # тестовый прием для кодирования подсказки
    base_text = inspect.cleandoc("""
    #!/usr/bin/env python
    # -*- coding: cp932 -*-
    pass  # \u3088\u308d\u3057\u304f
    """)

    pages = FilePaginator(BytesIO(base_text.encode("cp932"))).pages

    assert len(pages) == 1
    assert pages[0] == f"```python\n{base_text}\n```"

    # тест без кодирования подсказки
    with pytest.raises(UnicodeDecodeError):
        FilePaginator(BytesIO("\u3088\u308d\u3057\u304f".encode("cp932")))

    # Проверка с неправильным подсказкой
    with pytest.raises(UnicodeDecodeError):
        FilePaginator(BytesIO("-*- coding: utf-8 -*-\n\u3088\u308d\u3057\u304f".encode("cp932")))

    # Проверка ооб
    with pytest.raises(ValueError):
        FilePaginator(BytesIO("one\ntwo\nthree\nfour".encode('utf-8')), line_span=(-1, 20))


def test_wrapped_paginator():
    paginator = WrappedPaginator(max_size=200)
    paginator.add_line("abcde " * 50)
    assert len(paginator.pages) == 2

    paginator = WrappedPaginator(max_size=200, include_wrapped=False)
    paginator.add_line("abcde " * 50)
    assert len(paginator.pages) == 2


@pytest.mark.skipif(
    disnake.version_info >= (2, 0, 0),
    reason="Тесты с реакционной моделью границы раздела Paginator"
)
@utils.run_async
async def test_paginator_interface_reactions():
    bot = commands.Bot('?')

    with open(__file__, 'rb') as file:
        paginator = FilePaginator(file, max_size=200)

    interface = PaginatorInterface(bot, paginator)

    assert interface.pages == paginator.pages
    assert interface.page_count == len(paginator.pages)

    assert interface.page_size > 200
    assert interface.page_size < interface.max_page_size

    send_kwargs = interface.send_kwargs

    assert isinstance(send_kwargs, dict)
    assert 'content' in send_kwargs

    content = send_kwargs['content']

    assert isinstance(content, str)
    assert len(content) <= interface.page_size

    assert interface.display_page == 0

    # Страницы были закрыты, поэтому добавление строки должно сделать новую страницу
    old_page_count = interface.page_count

    await interface.add_line('a' * 150)

    assert interface.page_count > old_page_count

    # подтолкнуть страницу до конца (округлено на границы)
    interface.display_page = 999
    old_display_page = interface.display_page

    assert interface.pages == paginator.pages

    # Страница закрыта, поэтому создайте новую страницу
    await interface.add_line('b' * 150)

    # Убедитесь, что страница следовала за хвостом
    assert interface.display_page > old_display_page

    # Тестирование с помощью интерфейса встроенного
    embed_interface = PaginatorEmbedInterface(bot, paginator)

    assert embed_interface.pages[0] == interface.pages[0]

    send_kwargs = embed_interface.send_kwargs

    assert isinstance(send_kwargs, dict)
    assert 'embed' in send_kwargs

    embed = send_kwargs['embed']

    assert isinstance(embed, disnake.Embed)

    description = embed.description

    assert content.startswith(description)

    # Проверьте, чтобы получить повышение слишком большого размера страницы
    with pytest.raises(ValueError):
        PaginatorInterface(None, commands.Paginator(max_size=2000))

    # Проверьте на повышение на не-пагинаторе
    with pytest.raises(TypeError):
        PaginatorInterface(None, 4)

    paginator = commands.Paginator(max_size=100)
    for _ in range(100):
        paginator.add_line("test text")

    # тест взаимодействия
    with utils.mock_ctx(bot) as ctx:
        interface = PaginatorInterface(bot, paginator)

        assert not interface.closed

        await interface.send_to(ctx)

        await asyncio.sleep(0.1)
        await interface.add_line("test text")

        assert interface.page_count > 1
        assert not interface.closed

        interface.message.id = utils.sentinel()

        current_page = interface.display_page

        payload = {
            'message_id': interface.message.id,
            'user_id': ctx.author.id,
            'channel_id': ctx.channel.id,
            'guild_id': ctx.guild.id
        }

        # Нажмите правую кнопку
        emoji = disnake.PartialEmoji(
            animated=False,
            name="\N{BLACK RIGHT-POINTING TRIANGLE}",
            id=None
        )
        bot.dispatch(
            'raw_reaction_add',
            disnake.RawReactionActionEvent(payload, emoji, 'REACTION_ADD')
            if disnake.version_info >= (1, 3) else
            disnake.RawReactionActionEvent(payload, emoji)
        )

        await asyncio.sleep(0.1)

        assert interface.display_page > current_page
        assert not interface.closed

        current_page = interface.display_page

        # Нажмите на левую кнопку
        emoji = disnake.PartialEmoji(
            animated=False,
            name="\N{BLACK LEFT-POINTING TRIANGLE}",
            id=None
        )
        bot.dispatch(
            'raw_reaction_add',
            disnake.RawReactionActionEvent(payload, emoji, 'REACTION_ADD')
            if disnake.version_info >= (1, 3) else
            disnake.RawReactionActionEvent(payload, emoji)
        )

        await asyncio.sleep(0.1)

        assert interface.display_page < current_page
        assert not interface.closed

        # Нажмите кнопку последней страницы
        emoji = disnake.PartialEmoji(
            animated=False,
            name="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
            id=None
        )
        bot.dispatch(
            'raw_reaction_add',
            disnake.RawReactionActionEvent(payload, emoji, 'REACTION_ADD')
            if disnake.version_info >= (1, 3) else
            disnake.RawReactionActionEvent(payload, emoji)
        )

        await asyncio.sleep(0.1)

        assert interface.display_page == interface.page_count - 1
        assert not interface.closed

        # Нажмите кнопку первой страницы
        emoji = disnake.PartialEmoji(
            animated=False,
            name="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
            id=None
        )
        bot.dispatch(
            'raw_reaction_add',
            disnake.RawReactionActionEvent(payload, emoji, 'REACTION_ADD')
            if disnake.version_info >= (1, 3) else
            disnake.RawReactionActionEvent(payload, emoji)
        )

        await asyncio.sleep(0.1)

        assert interface.display_page == 0
        assert not interface.closed

        # Нажмите кнопку закрытия
        emoji = disnake.PartialEmoji(
            animated=False,
            name="\N{BLACK SQUARE FOR STOP}",
            id=None
        )
        bot.dispatch(
            'raw_reaction_add',
            disnake.RawReactionActionEvent(payload, emoji, 'REACTION_ADD')
            if disnake.version_info >= (1, 3) else
            disnake.RawReactionActionEvent(payload, emoji)
        )

        await asyncio.sleep(0.1)

        assert interface.closed
        ctx.send.coro.return_value.delete.assert_called_once()

    # тестировать повторную работу, без удаления
    with utils.mock_ctx(bot) as ctx:
        interface = PaginatorInterface(bot, paginator)

        assert not interface.closed

        await interface.send_to(ctx)

        await asyncio.sleep(0.1)
        await interface.add_line("test text")

        assert interface.page_count > 1
        assert not interface.closed

        # приезжать
        await interface.send_to(ctx)

        await asyncio.sleep(0.1)
        await interface.add_line("test text")

        ctx.send.coro.return_value.delete.assert_not_called()

        interface.task.cancel()
        await asyncio.sleep(0.1)

        assert interface.closed

    # Тестирование повторно, удалите
    with utils.mock_ctx(bot) as ctx:
        interface = PaginatorInterface(bot, paginator, delete_message=True)

        assert not interface.closed

        await interface.send_to(ctx)

        await asyncio.sleep(0.1)
        await interface.add_line("test text")

        assert interface.page_count > 1
        assert not interface.closed

        # приезжать
        await interface.send_to(ctx)

        await asyncio.sleep(0.1)
        await interface.add_line("test text")

        ctx.send.coro.return_value.delete.assert_called_once()

        interface.task.cancel()
        await asyncio.sleep(0.1)

        assert interface.closed

# TODO: Написать тест на интерфейс на основе взаимодействий на основе взаимодействий



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import copy

import disnake
from disnake.ext import commands


async def copy_context_with(ctx: commands.Context, *, author=None, channel=None, **kwargs):
    """
    Makes a new :class:`Context` with changed message properties.
    """

    # Скопируйте сообщение и обновите атрибуты
    alt_message: disnake.Message = copy.copy(ctx.message)
    alt_message._update(kwargs)

    if author is not None:
        alt_message.author = author
    if channel is not None:
        alt_message.channel = channel

    # Получить и вернуть контекст того же типа
    return await ctx.bot.get_context(alt_message, cls=type(ctx))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

from disnake.ext import commands

from jishaku.codeblocks import Codeblock, codeblock_converter
from jishaku.exception_handling import ReplResponseReactor
from jishaku.features.baseclass import Feature
from jishaku.paginators import PaginatorInterface, WrappedPaginator
from jishaku.shell import ShellReader


class ShellFeature(Feature):
    """
    Функция, содержащая команды, связанные с оболочкой
    """

    @Feature.Command(parent="jsk", name="shell", aliases=["bash", "sh", "powershell", "ps1", "ps", "cmd"])
    async def jsk_shell(self, ctx: commands.Context, *, argument: codeblock_converter):
        """
        Выполняет операторы в системной оболочке.

        Это использует системную оболочку, как определено в $ Shell, или `/bin/bash` в противном случае.
        Выполнение может быть отменено путем закрытия пагинатора.
        """

        async with ReplResponseReactor(ctx.message):
            with self.submit(ctx):
                with ShellReader(argument.content) as reader:
                    prefix = "```" + reader.highlight

                    paginator = WrappedPaginator(prefix=prefix, max_size=1975)
                    paginator.add_line(f"{reader.ps1} {argument.content}\n")

                    interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
                    self.bot.loop.create_task(interface.send_to(ctx))

                    async for line in reader:
                        if interface.closed:
                            return
                        await interface.add_line(line)

                await interface.add_line(f"\n[status] Return code {reader.close_code}")

    @Feature.Command(parent="jsk", name="git")
    async def jsk_git(self, ctx: commands.Context, *, argument: codeblock_converter):
        """
        Ярлык для «JSK SH GIT». Вызывает системную оболочку.
        """

        return await ctx.invoke(self.jsk_shell, argument=Codeblock(argument.language, "git " + argument.content))

    @Feature.Command(parent="jsk", name="pip")
    async def jsk_pip(self, ctx: commands.Context, *, argument: codeblock_converter):
        """
        Ярлык для «JSK SH Pip». Вызывает системную оболочку.
        """

        return await ctx.invoke(self.jsk_shell, argument=Codeblock(argument.language, "pip " + argument.content))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import asyncio
import functools
import typing


def executor_function(sync_function: typing.Callable):
    """Декоратор, который окутает функцию синхронизации в исполнителе, превращая ее в асинхронную функцию.

    Это позволяет обернуть и использовать и использовать функции обработки в качестве асинхронной функции.

    Примеры
    ---------

    Проталкивание обработки с помощью библиотеки изображений Python в исполнителя:

    .. code-block:: python3

        from io import BytesIO
        from PIL import Image

        from jishaku.functools import executor_function


        @executor_function
        def color_processing(color: disnake.Color):
            with Image.new('RGB', (64, 64), color.to_rgb()) as im:
                buff = BytesIO()
                im.save(buff, 'png')

            buff.seek(0)
            return buff

        @bot.command()
        async def color(ctx: commands.Context, color: disnake.Color=None):
            color = color or ctx.author.color
            buff = await color_processing(color=color)

            await ctx.send(file=disnake.File(fp=buff, filename='color.png'))
    """

    @functools.wraps(sync_function)
    async def sync_wrapper(*args, **kwargs):
        """
        Асинхронная функция, которая завершает функцию синхронизации с исполнителем.
        """

        loop = asyncio.get_event_loop()
        internal_function = functools.partial(sync_function, *args, **kwargs)
        return await loop.run_in_executor(None, internal_function)

    return sync_wrapper


class AsyncSender:
    """
    Класс потока хранения и управления, который позволяет более красиво отправлять в асинхронные итераторы.

    Пример
    --------

    .. code:: python3

        async def foo():
            print("foo yielding 1")
            x = yield 1
            print(f"foo received {x}")
            yield 3

        async for send, result in AsyncSender(foo()):
            print(f"asyncsender received {result}")
            send(2)

    Produces:

    .. code::

        foo yielding 1
        asyncsender received 1
        foo received 2
        asyncsender received 3
    """

    __slots__ = ('iterator', 'send_value')

    def __init__(self, iterator):
        self.iterator = iterator
        self.send_value = None

    def __aiter__(self):
        return self._internal(self.iterator.__aiter__())

    async def _internal(self, base):
        try:
            while True:
                # Send the last value to the iterator
                value = await base.asend(self.send_value)
                # Reset it incase one is not sent next iteration
                self.send_value = None
                # Yield sender and iterator value
                yield self.set_send_value, value
        except StopAsyncIteration:
            pass

    def set_send_value(self, value):
        """
        Устанавливает следующее значение, которое будет отправлено в итератор.

        Это обеспечивается итерацией этого класса и должно
        не называться напрямую.
        """

        self.send_value = value



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import collections

# Настройки смайликов, это устанавливает то, что смайлики используются для PaginatorInterface
EmojiSettings = collections.namedtuple('EmojiSettings', 'start back forward end close')

EMOJI_DEFAULT = EmojiSettings(
    start="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
    back="\N{BLACK LEFT-POINTING TRIANGLE}",
    forward="\N{BLACK RIGHT-POINTING TRIANGLE}",
    end="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}",
    close="\N{BLACK SQUARE FOR STOP}"
)



# -*- coding: utf-8 -*-

"""
jishaku converter test
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import inspect

from jishaku.codeblocks import Codeblock, codeblock_converter


def test_codeblock_converter():
    text = """
    ```py
    one
    ```
    """

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'one'
    assert codeblock.language == 'py'

    text = """
    ```sql
    two
    ```
    """

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'two'
    assert codeblock.language == 'sql'

    text = """
    ```txt
    three
    ```
    """

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'three'
    assert codeblock.language == 'txt'

    text = """
    ```
    four
    ```
    """

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'four'
    assert not codeblock.language

    text = "five"

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'five'
    assert not codeblock.language

    text = """
    six
    ```
    """

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'six\n```'
    assert not codeblock.language

    text = ""

    codeblock = codeblock_converter(inspect.cleandoc(text))

    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == ''
    assert not codeblock.language

    text = """
    ```
    se``ven
    ```
    """

    codeblock = codeblock_converter(inspect.cleandoc(text))
    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'se``ven'
    assert not codeblock.language

    text = "``ei`ght``"
    codeblock = codeblock_converter(text)
    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'ei`ght'
    assert not codeblock.language

    text = "`nine`"
    codeblock = codeblock_converter(text)
    assert isinstance(codeblock, Codeblock)
    assert codeblock.content.strip() == 'nine'
    assert not codeblock.language



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import inspect
import typing


class Scope:
    """
    Класс, который представляет глобальный и локальный объем как для осмотра, так и для творения.

    Многие функции Repl ожидают или возвращают этот класс.

    .. code:: python3

        scope = Scope()  # пустая область прицела

        scope = Scope(globals(), locals())  # Область, имитирующая текущую, реальную область.

        scope = Scope({'a': 3})  # Применение с ранее существовавшим глобальным ключом по объему и пустой локальной областью.
    """

    __slots__ = ('globals', 'locals')

    def __init__(self, globals_: dict = None, locals_: dict = None):
        self.globals: dict = globals_ or {}
        self.locals: dict = locals_ or {}

    def clear_intersection(self, other_dict):
        """
        Очищает местных жителей и глобалов из этой области, где совпадает пара клавишных значений
        с другим_диктом.

        Это позволяет очистить временные переменные, которые могли промыться в это
        Объем.

        Параметры
        -----------
        Другое_Дикт: :class:`dict`
            Словарь, который будет использоваться для определения очистки объема.

            Если ключ из этого дикта соответствует записи в глобальных или местных жителях этой области,
            и значение идентично, оно удаляется из области.

        Возврат
        -------
        Объем
            Обновленная область (я).
        """

        for key, value in other_dict.items():
            if key in self.globals and self.globals[key] is value:
                del self.globals[key]
            if key in self.locals and self.locals[key] is value:
                del self.locals[key]

        return self

    def update(self, other):
        """
        Обновляет эту область с помощью содержания другой области.

        Parameters
        ---------
        other: :class:`Scope`
            Область наложения на это.
        Возврат
        -------
        Объем
            Обновленная область (я).
        """

        self.globals.update(other.globals)
        self.locals.update(other.locals)
        return self

    def update_globals(self, other: dict):
        """
        Обновляют глобалы этой области диктом.

        Параметры
        -----------
        другой: :class:`dict`
            Словарь, который должен быть объединен в эту область.

        Возврат
        -------
        Объем
            Обновленная область (я).
        """

        self.globals.update(other)
        return self

    def update_locals(self, other: dict):
        """
        Обновляет местных жителей этой области диктом.

        Параметры
        -----------
        другой: :class:`dict`
            Словарь, который должен быть объединен в эту область.

        Возврат
        -------
        Объем
            Обновленная область (я).
        """

        self.locals.update(other)
        return self


def get_parent_scope_from_var(name, global_ok=False, skip_frames=0) -> typing.Optional[Scope]:
    """
    Отражает стек кадров в поисках кадра, содержащей заданное имя переменной.

    Возврат
    --------
    Необязательно [Scope]
        Соответствующий: класс: `recope` или нет
    """

    stack = inspect.stack()
    try:
        for frame_info in stack[skip_frames + 1:]:
            frame = None

            try:
                frame = frame_info.frame

                if name in frame.f_locals or (global_ok and name in frame.f_globals):
                    return Scope(globals_=frame.f_globals, locals_=frame.f_locals)
            finally:
                del frame
    finally:
        del stack

    return None


def get_parent_var(name, global_ok=False, default=None, skip_frames=0):
    """
    Непосредственно получает переменную от родительской кадры.

    Возврат
    --------
    Любой
        Содержание переменной, найденной данным именем, или нет.
    """

    scope = get_parent_scope_from_var(name, global_ok=global_ok, skip_frames=skip_frames + 1)

    if not scope:
        return default

    if name in scope.locals:
        return scope.locals.get(name, default)

    return scope.globals.get(name, default)



# -*- coding: utf-8 -*-

"""
jishaku.models tests
~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import utils

from jishaku.models import copy_context_with


@utils.run_async
async def test_context_copy():
    with utils.mock_ctx() as ctx:
        await copy_context_with(ctx, author=1, channel=2, content=3)

        ctx.bot.get_context.assert_called_once()
        alt_message = ctx.bot.get_context.call_args[0][0]

        alt_message._update.assert_called_once()
        assert alt_message._update.call_args[0] == ({"content": 3},)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import logging
import sys

import click
from disnake.ext import commands

LOG_FORMAT: logging.Formatter = logging.Formatter('%(asctime)s:%(levelname)s:%(name)s: %(message)s')
LOG_STREAM: logging.Handler = logging.StreamHandler(stream=sys.stdout)
LOG_STREAM.setFormatter(LOG_FORMAT)


@click.command()
@click.argument('token')
def entrypoint(token: str):
    """
    Входная точка доступна через `python -m jishaku <TOKEN>`
    """

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    logger.addHandler(LOG_STREAM)

    bot = commands.Bot(commands.when_mentioned)
    bot.load_extension('jishaku')
    bot.run(token)


if __name__ == '__main__':
    entrypoint()



# -*- coding: utf-8 -*-

"""
jishaku.repl internal test
~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import inspect
import random
import sys

import pytest
from utils import mock_ctx, run_async

from jishaku.repl import AsyncCodeExecutor, Scope, get_parent_var, get_var_dict_from_ctx


def upper_method():
    return get_parent_var('hidden_variable')


async def add_numbers(one, two):
    return one + two


@pytest.fixture(scope="module")
def scope():
    return Scope(
        {
            "add_numbers": add_numbers,
            "placement": 81
        },
        {
            "placement_local": 18
        }
    )


def test_scope_var():
    for _ in range(10):
        hidden_variable = random.randint(0, 1000000)
        test = upper_method()

        assert hidden_variable == test

        del hidden_variable

        test = upper_method()
        assert test is None

        assert get_parent_var('pytest', global_ok=True) == pytest


@pytest.mark.parametrize(
    ("code", "expected"),
    [
        ('3 + 4', [7]),
        ('return 3 + 9', [12]),
        ('yield 30; yield 40', [30, 40]),
        ('yield 60; 70', [60, 70]),
        ('90; 100', [100]),
        pytest.param(
            'eval("""\n77 + 22\n""")', [99],
            marks=pytest.mark.skipif(
                sys.version_info < (3, 7),
                reason="3.6 Требуется функция брони, поэтому не может справиться с отступами"
            )
        )
    ]
)
@run_async
async def test_executor_basic(code, expected):
    return_data = []
    async for result in AsyncCodeExecutor(code):
        return_data.append(result)

    assert len(return_data) == len(expected)
    for a, b in zip(return_data, expected):
        assert a == b


@pytest.mark.parametrize(
    ("code", "expected", "arg_dict"),
    [
        ("", [None], None),
        ("# this is a comment", [None], None),
        ("b = 12 + 82", [None], None),
        ("b", [94], None),
        ("c = placement + 7; c", [88], None),
        (
            "_cool_data + _not_so_cool",
            [445],
            {
                '_cool_data': 45,
                '_not_so_cool': 400
            }
        ),
        pytest.param(
            "_cool_data", [45], None,
            marks=pytest.mark.xfail(raises=NameError, strict=True)
        ),
        ("await add_numbers(10, 12)", [22], None)
    ]
)
@run_async
async def test_executor_advanced(code, expected, arg_dict, scope):

    return_data = []
    async for result in AsyncCodeExecutor(code, scope, arg_dict=arg_dict):
        return_data.append(result)

    assert len(return_data) == len(expected)
    for a, b in zip(return_data, expected):
        assert a == b

    if arg_dict:
        scope.clear_intersection(arg_dict)


@run_async
async def test_scope_copy(scope):
    scope2 = Scope()
    scope2.update(scope)

    assert scope.globals == scope2.globals, "Проверка сфера областей глобалов скопирована"
    assert scope.locals == scope2.locals, "Проверка прицела местных жителей скопирована"

    insert_dict = {'e': 7}
    scope.update_locals(insert_dict)

    assert 'e' in scope.locals, "Проверка обновления местных жителей"
    assert 'e' not in scope2.locals, "Проверка областей клона. Местные жители не обновлены"

    scope.clear_intersection(insert_dict)

    assert 'e' not in scope.locals, "Проверка пересечения местных жителей очищено"

    scope.update_globals(insert_dict)

    assert 'e' in scope.globals, "Проверка обновленных глобальных областей"
    assert 'e' not in scope2.globals, "Проверка областей клонов Globals не обновлена"

    scope.clear_intersection(insert_dict)

    assert 'e' not in scope.globals, "Проверка перекрестка Globals очищено"


@run_async
async def test_executor_builtins(scope):
    codeblock = inspect.cleandoc("""
    def ensure_builtins():
        return ValueError
    """)

    return_data = []
    async for result in AsyncCodeExecutor(codeblock, scope):
        return_data.append(result)

    assert len(return_data) == 1
    assert return_data[0] is None

    assert 'ensure_builtins' in scope.globals, "Проверка функции остается определенной"
    assert callable(scope.globals['ensure_builtins']), "Проверка определенной"
    assert scope.globals['ensure_builtins']() == ValueError, "Проверка определенного возврата согласованной"


def test_var_dict(scope):
    with mock_ctx() as ctx:
        scope.update_globals(get_var_dict_from_ctx(ctx))

        assert scope.globals['_ctx'] is ctx
        assert scope.globals['_bot'] is ctx.bot
        assert scope.globals['_message'] is ctx.message



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import asyncio
import collections
import contextlib
import typing
from datetime import datetime, timezone

from disnake.ext import commands

__all__ = (
    'Feature',
    'CommandTask'
)


CommandTask = collections.namedtuple("CommandTask", "index ctx task")


class Feature(commands.Cog):
    """
    Базовый определяющий функции компонентов Jishaku Cog.
    """

    class Command:
        """
        Промежуточный класс для команд функций.
        Экземпляры этого класса будут преобразованы в commands.command или commands.

        : param Parent: Чем должна быть основана эта команда.
        : PARAM STANDALONE_OK: должна ли команда быть разрешена быть автономной, если ее родитель не найден.
        """

        def __init__(self, parent: str = None, standalone_ok: bool = False, **kwargs):
            self.parent: typing.Union[str, Feature.Command] = parent
            self.standalone_ok = standalone_ok
            self.kwargs = kwargs
            self.callback = None
            self.depth: int = 0
            self.has_children: bool = False

        def __call__(self, callback: typing.Callable):
            self.callback = callback
            return self

    load_time: datetime = datetime.utcnow().replace(tzinfo=timezone.utc)

    def __init__(self, *args, **kwargs):
        self.bot: commands.Bot = kwargs.pop('bot')
        self.start_time: datetime = datetime.utcnow().replace(tzinfo=timezone.utc)
        self.tasks = collections.deque()
        self.task_count: int = 0

        # Генерировать и прикрепить команды
        command_lookup = {}

        for kls in reversed(type(self).__mro__):
            for key, cmd in kls.__dict__.items():
                if isinstance(cmd, Feature.Command):
                    command_lookup[key] = cmd

        command_set = list(command_lookup.items())

        # Попробуйте связать каждую родительскую команду со своим родителем
        for key, cmd in command_set:
            cmd.parent_instance = None
            cmd.depth = 0

            if cmd.parent and isinstance(cmd.parent, str):
                if cmd.standalone_ok:
                    cmd.parent_instance = command_lookup.get(cmd.parent, None)
                else:
                    try:
                        cmd.parent_instance = command_lookup[cmd.parent]
                    except KeyError as exception:
                        raise RuntimeError(
                            f"Не удалось связать функциональную команду {ключ} с ее родительской командой {cmd.parent}"
                        ) from exception
            # Также поднимите, если какой -либо команды не хватает обратного вызова
            if cmd.callback is None:
                raise RuntimeError(f"В функциональной команде {key} отсутствует обратный вызов")

        # Назначьте глубину и has_children
        for key, cmd in command_set:
            parent = cmd.parent_instance
            # Повторяют родителей, увеличивая глубину, пока мы не достигнем вершины
            while parent:
                parent.has_children = True
                cmd.depth += 1
                parent = parent.parent_instance

        # Сортировать по глубине
        command_set.sort(key=lambda c: c[1].depth)
        association_map = {}

        self.feature_commands = {}

        for key, cmd in command_set:
            if cmd.parent:
                parent = association_map[cmd.parent_instance]
                command_type = parent.group if cmd.has_children else parent.command
            else:
                command_type = commands.group if cmd.has_children else commands.command

            association_map[cmd] = target_cmd = command_type(**cmd.kwargs)(cmd.callback)
            target_cmd.cog = self
            self.feature_commands[key] = target_cmd
            setattr(self, key, target_cmd)

        self.__cog_commands__ = (*self.__cog_commands__, *self.feature_commands.values())

        # Не думайте, что это много, но все равно инициирует.
        super().__init__(*args, **kwargs)

    async def cog_check(self, ctx: commands.Context):
        """
        Локальная проверка, делает все команды в полученных Cogs только владельца
        """

        if not await ctx.bot.is_owner(ctx.author):
            raise commands.NotOwner("Вы должны владеть этим ботом, чтобы использовать Jishaku.")
        return True

    @contextlib.contextmanager
    def submit(self, ctx: commands.Context):
        """
        Контекст-менеджер, который передает текущую задачу в список задач Джишаку
и удаляет его потом.

        Параметры
        -----------
        ctx: commands.Context
            Контекстный объект, используемый для получения информации об этой командной задаче.
        """

        self.task_count += 1

        try:
            current_task = asyncio.current_task()
        except RuntimeError:
            # asyncio.current_task не документирует, что он может повысить время выполнения, но это так.
            # Он распространяется из asyncio.get_running_loop (), так что это происходит, когда не работает цикл.
            # Неясно, является ли это регрессией или преднамеренным изменением, поскольку в 3.6
            # asyncio.task.current_task () только не вернул бы в этом случае.
            current_task = None

        cmdtask = CommandTask(self.task_count, ctx, current_task)

        self.tasks.append(cmdtask)

        try:
            yield cmdtask
        finally:
            if cmdtask in self.tasks:
                self.tasks.remove(cmdtask)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

from disnake.ext import commands

from jishaku.flags import Flags
from jishaku.hljs import get_language, guess_file_traits
from jishaku.shim.paginator_base import EmojiSettings

from jishaku.shim.paginator_200 import PaginatorEmbedInterface, PaginatorInterface

__all__ = ('EmojiSettings', 'PaginatorInterface', 'PaginatorEmbedInterface',
           'WrappedPaginator', 'FilePaginator', 'use_file_check')


class WrappedPaginator(commands.Paginator):
    """
    Плаватор, который позволяет автоматическому обертыванию линий, если они не подходят.

    Это полезно при страничном непредсказуемом выходе,
    так как это позволяет расщеплять линии на большие куски данных.

    Разделители приоритеты в порядке их кортежа.

    Параметры
    -----------
    обернуть: tuple
        Круп с обертывающими делимитерами.
    Включить: bool
        Должно ли включать разделитель в начале новой оберщенной линии.
    принудительнаяОбертка: bool
        Если это правда, линии будут разделены на их максимальные точки
        с любым предоставленным разделителем.
    """

    def __init__(self, *args, wrap_on=('\n', ' '), include_wrapped=True, force_wrap=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.wrap_on = wrap_on
        self.include_wrapped = include_wrapped
        self.force_wrap = force_wrap

    def add_line(self, line='', *, empty=False):
        true_max_size = self.max_size - self._prefix_len - self._suffix_len - 2
        original_length = len(line)

        while len(line) > true_max_size:
            search_string = line[0:true_max_size - 1]
            wrapped = False

            for delimiter in self.wrap_on:
                position = search_string.rfind(delimiter)

                if position > 0:
                    super().add_line(line[0:position], empty=empty)
                    wrapped = True

                    if self.include_wrapped:
                        line = line[position:]
                    else:
                        line = line[position + len(delimiter):]

                    break

            if not wrapped:
                if self.force_wrap:
                    super().add_line(line[0:true_max_size - 1])
                    line = line[true_max_size - 1:]
                else:
                    raise ValueError(
                        f"Line of length {original_length} had sequence of {len(line)} characters"
                        f" (max is {true_max_size}) that WrappedPaginator could not wrap with"
                        f" delimiters: {self.wrap_on}"
                    )

        super().add_line(line, empty=empty)


class FilePaginator(commands.Paginator):
    """
    Парень из кодовых блоков с синтаксисом, прочитанным из файла, подобного файлу.

    Parameters
    -----------
    fp
        Файл, подобный (реализации ``fp.read``) Чтобы прочитать данные для этого пагинтора от.
    линейный промежуток: Optional[Tuple[int, int]]
        LineSpan для чтения из файла.Если нет, читает весь файл.
    language_hints: Tuple[str]
        Кортеж из струн, который может намекнуть на язык этого файла.
        Это может включать в себя имена файлов, типов MIME или шебанга.
        Шебанг, присутствующий в фактическом файле, всегда будет приоритетным из -за этого.
    """

    def __init__(self, fp, line_span=None, language_hints=(), **kwargs):
        language = ''

        for hint in language_hints:
            language = get_language(hint)

            if language:
                break

        if not language:
            try:
                language = get_language(fp.name)
            except AttributeError:
                pass

        content, _, file_language = guess_file_traits(fp.read())

        language = file_language or language
        lines = content.split('\n')

        super().__init__(prefix=f'```{language}', suffix='```', **kwargs)

        if line_span:
            line_span = sorted(line_span)

            if min(line_span) < 1 or max(line_span) > len(lines):
                raise ValueError("Linespan goes out of bounds.")

            lines = lines[line_span[0] - 1:line_span[1]]

        for line in lines:
            self.add_line(line)


class WrappedFilePaginator(FilePaginator, WrappedPaginator):
    """
    Комбинация FilePaginator и обернутого пагинтора.
    Другими словами, FilePaginator, который поддерживает обертывание линии.
    """


def use_file_check(ctx: commands.Context, size: int) -> bool:
    """
    Проверка, чтобы определить, является ли загрузка файла и полагаться на предварительный просмотр файла Discord, приемлемо по сравнению с PaginatorInterface.
    """

    return all([
        size < 50_000,  # Проверьте, что текст находится под точкой отсечения раздора;
        not Flags.FORCE_PAGINATOR,  # Проверьте, что пользователь явно не отключил это;
        (not ctx.author.is_on_mobile() if ctx.guild and ctx.bot.intents.presences else True)  # Убедитесь, что пользователь не на мобильном
    ])



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import typing

import disnake
import disnake.opus
import disnake.voice_client
from disnake.ext import commands

from jishaku.features.baseclass import Feature


class VoiceFeature(Feature):
    """
    Функция, содержащая основные команды, связанные с голосом
    """

    @staticmethod
    async def voice_check(ctx: commands.Context):
        """
        Проверьте, доступен ли VC в этом боте.
        """

        if not disnake.voice_client.has_nacl:
            return await ctx.send("Голос не может быть использован, так как PyNaCl не загружен.")

        if not disnake.opus.is_loaded():
            if hasattr(disnake.opus, '_load_default'):
                if not disnake.opus._load_default():
                    return await ctx.send(
                        "Невозможно использовать Voice, поскольку libopus не загружен, а попытка загрузить значение по умолчанию завершилась неудачей."
                    )
            else:
                return await ctx.send("Голос не может быть использован, так как libopus не загружен.")

    @staticmethod
    async def connected_check(ctx: commands.Context):
        """
        Проверьте, связаны ли мы с VC в этой гильдии.
        """

        voice = ctx.guild.voice_client

        if not voice or not voice.is_connected():
            return await ctx.send("Не подключен к голосовому каналу в этой гильдии. ")

    @staticmethod
    async def playing_check(ctx: commands.Context):
        """
        Проверяет, играем ли мы аудио в VC в этой гильдии.

        Это удваивается как проверка соединения.
        """

        check = await VoiceFeature.connected_check(ctx)
        if check:
            return check

        if not ctx.guild.voice_client.is_playing():
            return await ctx.send("Голосовой клиент в этой гильдии ни во что не играет.")

    @Feature.Command(parent="jsk", name="voice", aliases=["vc"],
                     invoke_without_command=True, ignore_extra=False)
    async def jsk_voice(self, ctx: commands.Context):
        """
        Связанные с голосом команды.

        Если призван без подкоманда, передает текущее состояние голоса.
        """

        if await self.voice_check(ctx):
            return

        # Дайте информацию о текущем голосовом клиенте, если есть один
        voice = ctx.guild.voice_client

        if not voice or not voice.is_connected():
            return await ctx.send("Не подключен.")

        await ctx.send(f"Подключен к {voice.channel.name}, "
                       f"{'остановлен' if voice.is_paused() else 'играет' if voice.is_playing() else 'праздный'}.")

    @Feature.Command(parent="jsk_voice", name="join", aliases=["connect"])
    async def jsk_vc_join(self, ctx: commands.Context, *,
                          destination: typing.Union[disnake.VoiceChannel, disnake.Member] = None):
        """
        Соединяет голосовой канал или перемещается к нему, если уже подключен.

        Прохождение голосового канала использует этот голосовой канал.
        Прохождение участника будет использовать текущий голосовой канал этого участника.
        Передача ничего не будет использовать голосовой канал автора.
        """

        if await self.voice_check(ctx):
            return

        destination = destination or ctx.author

        if isinstance(destination, disnake.Member):
            if destination.voice and destination.voice.channel:
                destination = destination.voice.channel
            else:
                return await ctx.send("У участника нет голосового канала.")

        voice = ctx.guild.voice_client

        if voice:
            await voice.move_to(destination)
        else:
            await destination.connect(reconnect=True)

        await ctx.send(f"Подключен к {destination.name}.")

    @Feature.Command(parent="jsk_voice", name="disconnect", aliases=["dc"])
    async def jsk_vc_disconnect(self, ctx: commands.Context):
        """
        Отключается от голосового канала в этой гильдии, если он есть.
        """

        if await self.connected_check(ctx):
            return

        voice = ctx.guild.voice_client

        await voice.disconnect()
        await ctx.send(f"Отключен от {voice.channel.name}.")

    @Feature.Command(parent="jsk_voice", name="stop")
    async def jsk_vc_stop(self, ctx: commands.Context):
        """
        Прекращает запуск аудио источника, если есть.
        """

        if await self.playing_check(ctx):
            return

        voice = ctx.guild.voice_client

        voice.stop()
        await ctx.send(f"Прекратилось воспроизведение звука в {voice.channel.name}.")

    @Feature.Command(parent="jsk_voice", name="pause")
    async def jsk_vc_pause(self, ctx: commands.Context):
        """
        Паузу используется источник аудио, если он есть.
        """

        if await self.playing_check(ctx):
            return

        voice = ctx.guild.voice_client

        if voice.is_paused():
            return await ctx.send("Звук уже поставлен на паузу.")

        voice.pause()
        await ctx.send(f"Приостановлено воспроизведение звука в {voice.channel.name}.")

    @Feature.Command(parent="jsk_voice", name="resume")
    async def jsk_vc_resume(self, ctx: commands.Context):
        """
        Возобновляет работающий аудио источник, если он есть.
        """

        if await self.playing_check(ctx):
            return

        voice = ctx.guild.voice_client

        if not voice.is_paused():
            return await ctx.send("Звук не на паузе.")

        voice.resume()
        await ctx.send(f"Возобновлено воспроизведение звука в {voice.channel.name}.")

    @Feature.Command(parent="jsk_voice", name="volume")
    async def jsk_vc_volume(self, ctx: commands.Context, *, percentage: float):
        """
        Регулирует объем аудио источника, если он поддерживается.
        """

        if await self.playing_check(ctx):
            return

        volume = max(0.0, min(1.0, percentage / 100))

        source = ctx.guild.voice_client.source

        if not isinstance(source, disnake.PCMVolumeTransformer):
            return await ctx.send("Этот источник не поддерживает регулировку громкости или "
                                  "интерфейс для этого не доступен.")

        source.volume = volume

        await ctx.send(f"Громкость установлена на {volume * 100:.2f}%")

    @Feature.Command(parent="jsk_voice", name="play", aliases=["play_local"])
    async def jsk_vc_play(self, ctx: commands.Context, *, uri: str):
        """
        Играет аудио напрямую из URI.

        Может быть либо локальный файл, либо аудио-ресурс в Интернете.
        """

        if await self.connected_check(ctx):
            return

        voice = ctx.guild.voice_client

        if voice.is_playing():
            voice.stop()

        # Удалить встроенные маскировщики, если они присутствуют
        uri = uri.lstrip("<").rstrip(">")

        voice.play(disnake.PCMVolumeTransformer(disnake.FFmpegPCMAudio(uri)))
        await ctx.send(f"Играет в {voice.channel.name}.")



# -*- coding: utf-8 -*-

"""
jishaku.hljs test
~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, see Лицензия для получения более подробной информации.

"""

import pytest

from jishaku.hljs import get_language


@pytest.mark.parametrize(
    ("filename", "language"),
    [
        ('base.py', 'py'),
        ('config.yml', 'yml'),
        ('requirements.txt', ''),
        ('#!/usr/bin/env python', 'python'),
        ('#!/usr/bin/unknown', '')
    ]
)
def test_hljs(filename, language):
    assert get_language(filename) == language



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import math
import sys
import typing
import os
import psutil
import time
import datetime
from datetime import timedelta

import disnake
from disnake.ext import commands

from jishaku.features.baseclass import Feature
from jishaku.flags import Flags, ENABLED_SYMBOLS
from jishaku.modules import package_version
from jishaku.paginators import PaginatorInterface

try:
    import psutil
except ImportError:
    psutil = None


def natural_size(size_in_bytes: int):
    """
    Преобразует несколько байтов в подходящую шкалевую единицу
Например.:
        1024 -> 1.00 KiB
        12345678 -> 11.77 MiB
    """
    units = ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')

    power = int(math.log(size_in_bytes, 1024))

    return f"{size_in_bytes / (1024 ** power):.2f} {units[power]}"


class RootCommand(Feature):
    """
    Функция, содержащая команду root jsk
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jsk.hidden = Flags.HIDE

    @Feature.Command(name="jishaku", aliases=["jsk"],
                     invoke_without_command=True, ignore_extra=False)
    async def jsk(self, ctx: commands.Context):
        """
        Джишаку отладки и диагностические команды.

        Эта команда сама по себе дает краткое изложение статуса.
        Вся другая функциональность находится в пределах его подкомандов.
        """

        summary = [
            f"disnake-jishaku-ru v{package_version('disnake-jishaku-ru')} Test, "
            f"disnake `{package_version('disnake')}`, "

            f"`Python {sys.version}` на `{sys.platform}`".replace("\n", ""),
            f"Модуль был загружен <t:{self.load_time.timestamp():.0f}:R>, "
            f"ког был загружен <t:{self.start_time.timestamp():.0f}:R>.",
            ""
        ]

        # обнаружить, установлена ​​ли функция [procinfo]
        if psutil:
            try:
                proc = psutil.Process()

                with proc.oneshot():
                    try:
                        mem = proc.memory_full_info()
                        summary.append(f"Используя {natural_size(mem.rss)} физической памяти и "
                                       f"{natural_size(mem.vms)} виртуальной памяти, "
                                       f"{natural_size(mem.uss)} из которых уникальны для этого процесса.")
                    except psutil.AccessDenied:
                        pass

                    try:
                        name = proc.name()
                        pid = proc.pid
                        thread_count = proc.num_threads()

                        summary.append(f"Запущен на PID {pid} (`{name}`) с {thread_count} потоком(ами).")
                    except psutil.AccessDenied:
                        pass

                    summary.append("")  # пустая линия
            except psutil.AccessDenied:
                summary.append(
                    "psutil установлен, но у этого процесса недостаточно высокие права доступа "
                    "для запроса информации о процессах."
                )
                summary.append("")  # пустая линия

        cache_summary = f"{len(self.bot.guilds)} сервера(ов) и {len(self.bot.users)} пользователь(ей)"

        # Показать настройки Shard для резюме
        if isinstance(self.bot, disnake.AutoShardedClient):
            if len(self.bot.shards) > 20:
                summary.append(
                    f"Этот бот автоматически разделен на шарды ({len(self.bot.shards)} шарды из {self.bot.shard_count})"
                    f" и может видеть {cache_summary}."
                )
            else:
                shard_ids = ', '.join(str(i) for i in self.bot.shards.keys())
                summary.append(
                    f"Этот бот автоматически разделен на шарды (Шарды {shard_ids} из {self.bot.shard_count})"
                    f" и может видеть {cache_summary}."
                )
        elif self.bot.shard_count:
            summary.append(
                f"Этот бот разделен на шарды вручную (Шард {self.bot.shard_id} из {self.bot.shard_count})"
                f" и может видеть {cache_summary}."
            )
        else:
            summary.append(f"Этот бот не разделен на шарды и может видеть {cache_summary}.")

        if self.bot._connection.max_messages:
            message_cache = f"Кэш сообщений ограничен (`{self.bot._connection.max_messages}`)"
        else:
            message_cache = "Кэш сообщений отключён."

        if disnake.version_info >= (1, 5, 0):
            presence_intent = f"Presence интент {'включён' if self.bot.intents.presences else 'выключен'}"
            members_intent = f"Members интент {'включён' if self.bot.intents.members else 'выключен'}"

            summary.append(f"{message_cache}.\n\n> - {presence_intent}.\n> - {members_intent}.\n")
        else:
            guild_subscriptions = f"Подписки на сервера {'включены' if self.bot._connection.guild_subscriptions else 'выключены'}"

            summary.append(f"{message_cache} и {guild_subscriptions}.")

        _embed = os.getenv('JISHAKU_EMBEDDED_JSK', None)
        if _embed in ENABLED_SYMBOLS:
            _color = os.getenv('JISHAKU_EMBEDDED_JSK_COLOR', None) or os.getenv('JISHAKU_EMBEDDED_JSK_COLOUR', None)
            if _color is not None:
                try:
                    color = await commands.ColourConverter().convert(ctx, _color)
                except commands.errors.BadColourArgument:
                    color = disnake.Colour.default()
            else:
                color = disnake.Colour.default()

            await ctx.send(embed=disnake.Embed(
                description="\n".join(summary),
                color=color)
                .set_author(
                    name='Jishaku',
                    url=ctx.bot.user.display_avatar.url,
                    icon_url=ctx.bot.user.display_avatar.url)
                .set_footer(text=f"Задержка вебсокета в среднем: {round(self.bot.latency * 1000, 2)}ms")
            )
        else:
            summary.append(f"Задержка вебсокета в среднем: {round(self.bot.latency * 1000, 2)}ms")
            await ctx.send("\n".join(summary))

    @Feature.Command(parent="jsk", name="hide")
    async def jsk_hide(self, ctx: commands.Context):
        """
        Hides Jishaku from the help command.
        """

        if self.jsk.hidden:
            return await ctx.send("Jishaku уже скрыт.")

        self.jsk.hidden = True
        await ctx.send("Jishaku теперь скрыт.")

    @Feature.Command(parent="jsk", name="show")
    async def jsk_show(self, ctx: commands.Context):
        """
        Показывает Джишаку в команде справки.
        """

        if not self.jsk.hidden:
            return await ctx.send("Jishaku уже виден.")

        self.jsk.hidden = False
        await ctx.send("Jishaku теперь виден.")

    @Feature.Command(parent="jsk", name="tasks")
    async def jsk_tasks(self, ctx: commands.Context):
        """
        Показывает в настоящее время бегущие задачи Джишаку.
        """

        if not self.tasks:
            return await ctx.send("В данный момент нет запущенных задач.")

        paginator = commands.Paginator(max_size=1985)

        for task in self.tasks:
            paginator.add_line(f"{task.index}: `{task.ctx.command.qualified_name}`, вызывается в "
                               f"{task.ctx.message.created_at.strftime('%Y-%m-%d %H:%M:%S')} UTC")

        interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
        return await interface.send_to(ctx)

    @Feature.Command(parent="jsk", name="cancel")
    async def jsk_cancel(self, ctx: commands.Context, *, index: typing.Union[int, str]):
        """
        Отменяет задачу с данным индексом.

        Если пройден индекс -1, вместо этого отменит последнюю задачу.
        """

        if not self.tasks:
            return await ctx.send("Нет задач для отмены.")

        if index == "~":
            task_count = len(self.tasks)

            for task in self.tasks:
                task.task.cancel()

            self.tasks.clear()

            return await ctx.send(f"Отменено {task_count} задач.")

        if isinstance(index, str):
            raise commands.BadArgument('Литерал для обозначения "индекса" не распознается.')

        if index == -1:
            task = self.tasks.pop()
        else:
            task = disnake.utils.get(self.tasks, index=index)
            if task:
                self.tasks.remove(task)
            else:
                return await ctx.send("Несуществующая задача.")

        task.task.cancel()
        return await ctx.send(f"Отменена задача {task.index}: `{task.ctx.command.qualified_name}`,"
                              f" вызывается в {task.ctx.message.created_at.strftime('%Y-%m-%d %H:%M:%S')} UTC")



import pathlib
import re

from setuptools import setup

ROOT = pathlib.Path(__file__).parent

with open(ROOT / 'jishaku' / 'meta.py', 'r', encoding='utf-8') as f:
    VERSION_MATCH = re.search(r'VersionInfo\(major=(\d+), minor=(\d+), micro=(\d+), .+\)', f.read(), re.MULTILINE)

    if not VERSION_MATCH:
        raise RuntimeError('Версия не установлена ​​или не может быть найдена')

    VERSION = '.'.join([VERSION_MATCH.group(1), VERSION_MATCH.group(2), VERSION_MATCH.group(3)])

EXTRA_REQUIRES = {}

for feature in (ROOT / 'requirements').glob('*.txt'):
    with open(feature, 'r', encoding='utf-8') as f:
        EXTRA_REQUIRES[feature.with_suffix('').name] = f.read().splitlines()

REQUIREMENTS = EXTRA_REQUIRES.pop('_', [])

if not VERSION:
    raise RuntimeError('Версия не установлена')

with open(ROOT / 'README.md', 'r', encoding='utf-8') as f:
    README = f.read()

setup(
    name='disnake-jishaku-ru',
    author='darkness800',
    url='https://github.com/darkness800/disnake-jishaku-ru',

    license='MIT',
    description='Расширение Disnake, включая полезные инструменты для разработки и отладки ботов.',
    long_description=README,
    long_description_content_type='text/markdown',
    project_urls={
        'Code': 'https://github.com/darkness800/disnake-jishaku-ru',
        'Issue tracker': 'https://github.com/darkness800/disnake-jishaku-ru/issues'
    },

    version=VERSION,
    packages=['jishaku', 'jishaku.features', 'jishaku.repl', 'jishaku.shim'],
    include_package_data=True,
    install_requires=REQUIREMENTS,
    python_requires='>=3.8.0',

    extras_require=EXTRA_REQUIRES,

    keywords='jishaku disnake disnake-jishaku-ru discord cog repl extension ru-fork',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Russian',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Communications :: Chat',
        'Topic :: Internet',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities'
    ]
)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

from collections import namedtuple

import pkg_resources

__all__ = (
    '__author__',
    '__copyright__',
    '__docformat__',
    '__license__',
    '__title__',
    '__version__',
    'version_info'
)

# pylint: disable=invalid-name
VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
version_info = VersionInfo(major=1, minor=5, micro=6, releaselevel='final', serial=0)

__author__ = 'Darkness800'
__copyright__ = 'Copyright 2021 Devon (Gorialis) R, Translated By darkness.py'
__docformat__ = 'restructuredtext ru'
__license__ = 'MIT'
__title__ = 'jishaku'
__version__ = '.'.join(map(str, (version_info.major, version_info.minor, version_info.micro)))

# Это гарантирует, что когда Джишаку перезагружен, pkg_resources требует, чтобы она предоставила правильную информацию о версии
pkg_resources.working_set.by_key.pop('jishaku', None)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import asyncio
import subprocess
import traceback
import typing

import disnake
from disnake.ext import commands

from jishaku.flags import Flags


async def send_traceback(destination: disnake.abc.Messageable, verbosity: int, *exc_info):
    """
    Отправляет трассировку исключения из пункта назначения.
    Используется, когда реплика не удается по любой причине.

    :param destination: Куда отправить эту информацию
    :param verbosity: Как далеко должен идти этот трассировка.0 показывает только последний стек.
    :param exc_info: Информация об этом исключении, от sys.exc_info или аналогичного.
    :return: Последнее сообщение отправлено
    """

    etype, value, trace = exc_info

    traceback_content = "".join(traceback.format_exception(etype, value, trace, verbosity)).replace("``", "`\u200b`")

    paginator = commands.Paginator(prefix='```py')
    for line in traceback_content.split('\n'):
        paginator.add_line(line)

    message = None

    for page in paginator.pages:
        message = await destination.send(page)

    return message


async def do_after_sleep(delay: float, coro, *args, **kwargs):
    """
    Выполняет действие после установленного количества времени.

    Эта функция вызывает кораку только после задержки,
    Предотвращение жалоб асинсио на разрушенные короны.

    :param delay: Время за считанные секунды
    :param coro: Coroutine для бега
    :param args: Аргументы, чтобы передать в Coroutine
    :param kwargs: Аргументы ключевого слова для перехода в Coroutine
    :return: Что бы ни вернулась коратика.
    """
    await asyncio.sleep(delay)
    return await coro(*args, **kwargs)


async def attempt_add_reaction(msg: disnake.Message, reaction: typing.Union[str, disnake.Emoji])\
        -> typing.Optional[disnake.Reaction]:
    """
    Попробуйте добавить реакцию на сообщение, игнорируя его, если оно не по какой -либо причине.
    :param msg: Сообщение, чтобы добавить реакцию.
    :param reaction: Реакция эмодзи, может быть строкой или `disnake.Emoji`
    :return: а `disnake.Reaction` или нет, в зависимости от того, не удалось или нет.
    """
    try:
        return await msg.add_reaction(reaction)
    except disnake.HTTPException:
        pass


class ReactionProcedureTimer:
    """
    Класс, который реагирует на сообщение, основанное на том, что происходит в течение его жизни.
    """
    __slots__ = ('message', 'loop', 'handle', 'raised')

    def __init__(self, message: disnake.Message, loop: typing.Optional[asyncio.BaseEventLoop] = None):
        self.message = message
        self.loop = loop or asyncio.get_event_loop()
        self.handle = None
        self.raised = False

    async def __aenter__(self):
        self.handle = self.loop.create_task(do_after_sleep(1, attempt_add_reaction, self.message,
                                                           "\N{BLACK RIGHT-POINTING TRIANGLE}"))
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.handle:
            self.handle.cancel()

        # Нет исключения, отметка
        if not exc_val:
            await attempt_add_reaction(self.message, "\N{WHITE HEAVY CHECK MARK}")
            return

        self.raised = True

        if isinstance(exc_val, (asyncio.TimeoutError, subprocess.TimeoutExpired)):
            # Временный, будильник
            await attempt_add_reaction(self.message, "\N{ALARM CLOCK}")
        elif isinstance(exc_val, SyntaxError):
            # Ошибка синтаксиса, одиночная восклицательная отметка
            await attempt_add_reaction(self.message, "\N{HEAVY EXCLAMATION MARK SYMBOL}")
        else:
            # Другая ошибка, двойной восклицательный знак
            await attempt_add_reaction(self.message, "\N{DOUBLE EXCLAMATION MARK}")


class ReplResponseReactor(ReactionProcedureTimer):
    """
    Расширение реакционного процесса, которое поглощает ошибки, отправляя трассировки.
    """

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)

        # Ничто не пошло не так, кого это волнует, лол
        if not exc_val:
            return

        if isinstance(exc_val, (SyntaxError, asyncio.TimeoutError, subprocess.TimeoutExpired)):
            # Короткий след, отправить на канал
            await send_traceback(self.message.channel, 0, exc_type, exc_val, exc_tb)
        else:
            # Этот след, вероятно, нуждается в большей информации, поэтому увеличивайте многословие, и вместо этого DM.
            await send_traceback(
                self.message.channel if Flags.NO_DM_TRACEBACK else self.message.author,
                8, exc_type, exc_val, exc_tb
            )

        return True  # исключение было обработано



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import contextlib
import inspect
import io
import pathlib
import re
import time
import typing

import disnake
from disnake.ext import commands
from disnake.ext.commands.slash_core import (
    InvokableSlashCommand,
    SubCommandGroup,
    SubCommand
)

from jishaku.exception_handling import ReplResponseReactor
from jishaku.features.baseclass import Feature
from jishaku.models import copy_context_with
from jishaku.paginators import PaginatorInterface, WrappedPaginator, use_file_check

UserIDConverter = commands.IDConverter[disnake.User]


class SlimUserConverter(UserIDConverter):
    """
    Идентичный аккумулятору, но не выполняет проверки имен открытого текста.
    """

    async def convert(self, ctx: commands.Context, argument: str) -> disnake.User:
        """Метод преобразователя"""
        match = self._get_id_match(argument) or re.match(r'<@!?([0-9]{15,20})>$', argument)

        if match is not None:
            user_id = int(match.group(1))
            result = ctx.bot.get_user(user_id) or disnake.utils.get(ctx.message.mentions, id=user_id)
            if result is None:
                try:
                    result = await ctx.bot.fetch_user(user_id)
                except disnake.HTTPException:
                    raise commands.UserNotFound(argument) from None

            return result

        raise commands.UserNotFound(argument)


class InvocationFeature(Feature):
    """
    Функция, содержащая команды, связанные с вызовами команды
    """

    OVERRIDE_SIGNATURE = typing.Union[SlimUserConverter, disnake.TextChannel, disnake.Thread]

    @Feature.Command(parent="jsk", name="override", aliases=["execute", "exec", "override!", "execute!", "exec!"])
    async def jsk_override(
        self, ctx: commands.Context,
        overrides: commands.Greedy[OVERRIDE_SIGNATURE],
        *, command_string: str
    ):
        """
        Запустите команду с другим пользователем, каналом или потоком, опционально обойдя проверки и восстановления.

        Пользователи попытаются решить с участником, но будут использовать пользователя, если он не сможет его найти.
        """

        kwargs = {
            "content": ctx.prefix + command_string.lstrip('/')
        }

        for override in overrides:
            if isinstance(override, disnake.User):
                # Это пользователь
                if ctx.guild:
                    # Попробуйте перейти на экземпляр участника
                    # Раньше это делалось с помощью профсоюзного конвертера, но делать это так делает
                    #  Команда более совместима с цепочкой, например,`jsk in .. jsk su ..`
                    target_member = None

                    with contextlib.suppress(disnake.HTTPException):
                        target_member = ctx.guild.get_member(override.id) or await ctx.guild.fetch_member(override.id)

                    kwargs["author"] = target_member or override
                else:
                    kwargs["author"] = override
            else:
                # В противном случае это текстовый канал или поток
                kwargs["channel"] = override

        alt_ctx = await copy_context_with(ctx, **kwargs)

        if alt_ctx.command is None:
            if alt_ctx.invoked_with is None:
                return await ctx.send('Этот бот был жестко настроен на игнорирование этого пользователя.')
            return await ctx.send('Команда "{alt_ctx.invoked_with}" не существует.')

        if ctx.invoked_with.endswith('!'):
            return await alt_ctx.command.reinvoke(alt_ctx)

        return await alt_ctx.command.invoke(alt_ctx)

    @Feature.Command(parent="jsk", name="repeat")
    async def jsk_repeat(self, ctx: commands.Context, times: int, *, command_string: str):
        """
        Запускает команду несколько раз подряд.

        Это действует так, как команда была вызвана несколько раз вручную, так что это подчиняется усаживанию.
        Вы можете использовать это в сочетании с `JSK Sudo`, чтобы обойти это.
        """

        with self.submit(ctx):  # Разрешить повторения быть отменены
            for _ in range(times):
                alt_ctx = await copy_context_with(ctx, content=ctx.prefix + command_string)

                if alt_ctx.command is None:
                    return await ctx.send(f'Команда "{alt_ctx.invoked_with}" не существует.')

                await alt_ctx.command.reinvoke(alt_ctx)

    @Feature.Command(parent="jsk", name="debug", aliases=["dbg"])
    async def jsk_debug(self, ctx: commands.Context, *, command_string: str):
        """
        Запустите выполнение времени команды и улавливая исключения.
        """

        alt_ctx = await copy_context_with(ctx, content=ctx.prefix + command_string)

        if alt_ctx.command is None:
            return await ctx.send(f'Команда "{alt_ctx.invoked_with}" не существует.')

        start = time.perf_counter()

        async with ReplResponseReactor(ctx.message):
            with self.submit(ctx):
                await alt_ctx.command.invoke(alt_ctx)

        end = time.perf_counter()
        return await ctx.send(f"Команда `{alt_ctx.command.qualified_name}` выполнена за {end - start:.3f}сек.")

    def get_slash_command(
        self,
        name: str
    ) -> typing.Optional[typing.Union[InvokableSlashCommand, SubCommandGroup, SubCommand]]:
        """Работает как ``Bot.get_command``, Но для ударов.

        Параметры
        -----------
        имя: :class:`str`
            Имя команды Slash, чтобы получить.

        Возврат
        --------
        Optional[Union[:class:`InvokableSlashCommand`, :class:`SubCommandGroup`, :class:`SubCommand`]]
            Команда SLASH, которая была запрошена.Если не найдено, возвращается ``None``.
        """

        if not isinstance(name, str):
            raise TypeError(f"Ожидаемое имя должно быть str, а не {name.__class__}")

        command = name.split()
        slash = self.bot.get_slash_command(command[0])
        if slash:
            if len(command) == 1:
                return slash
            elif len(command) == 2:
                cmd = slash.children.get(command[1])
                return cmd
            elif len(command) == 3:
                group = slash.children.get(command[1])
                if isinstance(group, SubCommandGroup):
                    cmd = group.children.get(command[2])
                    return cmd

        return None

    @Feature.Command(parent="jsk", name="source", aliases=["src"])
    async def jsk_source(self, ctx: commands.Context, *, command_name: str):
        """
        Отображает исходный код для команды.
        """

        command = self.bot.get_command(command_name) or self.get_slash_command(command_name)
        if not command:
            return await ctx.send(f"Не удалось найти команду `{command_name}`.")

        try:
            source_lines, _ = inspect.getsourcelines(command.callback)
        except (TypeError, OSError):
            return await ctx.send(f"По какой-то причине не удалось получить исходный код для `{command}`.")

        filename = "source.py"

        try:
            filename = pathlib.Path(inspect.getfile(command.callback)).name
        except (TypeError, OSError):
            pass

        # Getourcelines по какой -то причине возвращается с линейными окончаниями
        source_text = ''.join(source_lines)

        if use_file_check(ctx, len(source_text)):  # Файл "Полный контент" Предел предварительного просмотра
            await ctx.send(file=disnake.File(
                filename=filename,
                fp=io.BytesIO(source_text.encode('utf-8'))
            ))
        else:
            paginator = WrappedPaginator(prefix='```py', suffix='```', max_size=1985)

            paginator.add_line(source_text.replace('```', '``\N{zero width space}`'))

            interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
            await interface.send_to(ctx)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import collections

__all__ = ('Codeblock', 'codeblock_converter')

Codeblock = collections.namedtuple('Codeblock', 'language content')


def codeblock_converter(argument):
    """
    Преобразователь, который лишает отметку кода, если он существует.

    Возвращает название (language, content).

    :attr:`Codeblock.language` это пустая строка, если с этим кодовым блоком не было дано.
    Это ``None`` Если ввод не был полным блоком кода.
    """
    if not argument.startswith('`'):
        return Codeblock(None, argument)

    # Держите небольшой буфер из последних очагов, которые мы видели
    last = collections.deque(maxlen=3)
    backticks = 0
    in_language = False
    in_code = False
    language = []
    code = []

    for char in argument:
        if char == '`' and not in_code and not in_language:
            backticks += 1  # Чтобы помочь отслеживать закрытие бэктиков
        if last and last[-1] == '`' and char != '`' or in_code and ''.join(last) != '`' * backticks:
            in_code = True
            code.append(char)
        if char == '\n':  # \ n делиминации язык и код
            in_language = False
            in_code = True
        # Мы еще не видим новую линию, но мы также прошли открытие ```
        elif ''.join(last) == '`' * 3 and char != '`':
            in_language = True
            language.append(char)
        elif in_language:  # Мы на языке после первого небакового персонажа
            if char != '\n':
                language.append(char)

        last.append(char)

    if not code and not language:
        code[:] = last

    return Codeblock(''.join(language), ''.join(code[len(language):-backticks]))



# -*- coding: utf-8 -*-

"""
jishaku.cog loadability and functionality test
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (Gorialis) R
:license: MIT, Смотрите лицензию для более подробной информации.

"""

import asyncio

import pytest
import utils
from disnake.ext import commands


@pytest.fixture(
    scope='module',
    params=[
        ("jishaku", commands.Bot, {}),
        ("jishaku", commands.Bot, {"shard_id": 0, "shard_count": 2}),
        ("jishaku", commands.AutoShardedBot, {}),
        ("jishaku.cog", commands.Bot, {}),
        ("jishaku.cog", commands.Bot, {"shard_id": 0, "shard_count": 2}),
        ("jishaku.cog", commands.AutoShardedBot, {}),
    ],
    ids=[
        "jishaku (Bot, unsharded)",
        "jishaku (Bot, sharded)",
        "jishaku (AutoShardedBot)",
        "jishaku.cog (Bot, unsharded)",
        "jishaku.cog (Bot, sharded)",
        "jishaku.cog (AutoShardedBot)"
    ]
)
def bot(request):
    b = request.param[1]('?', **request.param[2])
    b.load_extension(request.param[0])

    yield b

    b.unload_extension(request.param[0])
    b.loop.run_until_complete(b.close())


def test_loads(bot):
    assert bot.get_cog("Jishaku")
    assert isinstance(bot.get_cog("Jishaku"), commands.Cog)

    assert bot.get_command("jsk")
    assert isinstance(bot.get_command("jsk"), commands.Command)


def test_cog_attributes(bot):
    cog = bot.get_cog("Jishaku")

    cog.retain = False
    assert cog.scope is not cog.scope, "Собственность Scope должна дать новые прицелы на не удерживать"

    cog.retain = True
    assert cog.scope is cog.scope, "Свойство по объему должно быть последовательным при удержании"

    assert not cog.tasks

    with cog.submit("mock 1") as cmd_task:
        assert len(cog.tasks) == 1

        assert cmd_task.index == 1
        assert cmd_task.ctx == "mock 1"
        assert cmd_task.task is None

    assert not cog.tasks

    with cog.submit("mock 2") as cmd_task:
        assert len(cog.tasks) == 1

        assert cmd_task.index == 2
        assert cmd_task.ctx == "mock 2"
        assert cmd_task.task is None

    assert not cog.tasks


@utils.run_async
async def test_cog_check(bot):
    cog = bot.get_cog("Jishaku")

    with utils.mock_ctx() as ctx:
        with utils.mock_coro(ctx.bot, 'is_owner'):
            ctx.bot.is_owner.coro.return_value = True

            assert await cog.cog_check(ctx)

            ctx.bot.is_owner.coro.return_value = False

            with pytest.raises(commands.NotOwner):
                await cog.cog_check(ctx)


@utils.run_async
async def test_commands(bot):
    cog = bot.get_cog("Jishaku")

    # test 'jsk'
    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk').callback(cog, ctx)

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "Module was loaded" in text

    # test 'jsk hide' and 'jsk show'
    cog.jsk.hidden = False

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk hide').callback(cog, ctx)

        assert cog.jsk.hidden

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "now hidden" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk hide').callback(cog, ctx)

        assert cog.jsk.hidden

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "already hidden" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk show').callback(cog, ctx)

        assert not cog.jsk.hidden

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "now visible" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk show').callback(cog, ctx)

        assert not cog.jsk.hidden

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "already visible" in text

    # тест 'jsk tasks'
    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk tasks').callback(cog, ctx)

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "No currently running tasks" in text

    with utils.mock_ctx() as ctx:
        with cog.submit(ctx):
            interface = await bot.get_command('jsk tasks').callback(cog, ctx)

            ctx.send.assert_called_once()

            interface.task.cancel()

    # тест 'jsk cancel'
    with utils.mock_ctx() as ctx:
        # test explicit
        with cog.submit(ctx) as command_task:
            with pytest.raises(asyncio.CancelledError):
                await bot.get_command('jsk cancel').callback(cog, ctx, index=command_task.index)
                await asyncio.sleep(0.1)

            ctx.send.assert_called_once()
            text = ctx.send.call_args[0][0]
            assert f"Cancelled task {command_task.index}" in text

    with utils.mock_ctx() as ctx:
        # тест implicit
        with cog.submit(ctx) as command_task:
            with pytest.raises(asyncio.CancelledError):
                await bot.get_command('jsk cancel').callback(cog, ctx, index=-1)
                await asyncio.sleep(0.1)

            ctx.send.assert_called_once()
            text = ctx.send.call_args[0][0]
            assert f"Cancelled task {command_task.index}" in text

    with utils.mock_ctx() as ctx:
        # тест unknown task
        with cog.submit(ctx) as command_task:
            await bot.get_command('jsk cancel').callback(cog, ctx, index=123456789012345678)

            ctx.send.assert_called_once()
            text = ctx.send.call_args[0][0]
            assert "Unknown task" in text

    with utils.mock_ctx() as ctx:
        # тест no tasks
        await bot.get_command('jsk cancel').callback(cog, ctx, index=123456789012345678)

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "No tasks" in text

    # тест 'jsk retain'
    cog.retain = False

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk retain').callback(cog, ctx, toggle=True)

        assert cog.retain

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "is ON" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk retain').callback(cog, ctx, toggle=True)

        assert cog.retain

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "уже установлен ON" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk retain').callback(cog, ctx, toggle=None)

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "установлен на ON" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk retain').callback(cog, ctx, toggle=False)

        assert not cog.retain

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "является OFF" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk retain').callback(cog, ctx, toggle=False)

        assert not cog.retain

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "уже установлен OFF" in text

    with utils.mock_ctx() as ctx:
        await bot.get_command('jsk retain').callback(cog, ctx, toggle=None)

        ctx.send.assert_called_once()
        text = ctx.send.call_args[0][0]
        assert "установлен на OFF" in text



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

import itertools
import math
import time
import traceback
from urllib.parse import urlencode
import os
import sys

import disnake
from disnake.ext import commands

from jishaku.features.baseclass import Feature
from jishaku.flags import Flags
from jishaku.modules import ExtensionConverter
from jishaku.paginators import WrappedPaginator


class ManagementFeature(Feature):
    """
    Функция, содержащая команды управления расширением и бота
    """

    @Feature.Command(parent="jsk", name="load", aliases=["reload"])
    async def jsk_load(self, ctx: commands.Context, *extensions: ExtensionConverter):
        """
        Загружает или перезагружает заданные имена расширения.

        Сообщает о любых расширениях, которые не загружались.
        """

        paginator = WrappedPaginator(prefix='', suffix='')

        # 'JSK RELOAD' Сама только что перезагружает Джишаку
        if ctx.invoked_with == 'reload' and not extensions:
            extensions = [['jishaku']]

        for extension in itertools.chain(*extensions):
            method, icon = (
                (self.bot.reload_extension, "\N{CLOCKWISE RIGHTWARDS AND LEFTWARDS OPEN CIRCLE ARROWS}")
                if extension in self.bot.extensions else
                (self.bot.load_extension, "\N{INBOX TRAY}")
            )

            try:
                method(extension)
            except Exception as exc:
                traceback_data = ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__, 1))

                paginator.add_line(
                    f"{icon}\N{WARNING SIGN} `{extension}`\n```py\n{traceback_data}\n```",
                    empty=True
                )
            else:
                paginator.add_line(f"{icon} `{extension}`", empty=True)

        for page in paginator.pages:
            await ctx.send(page)

    @Feature.Command(parent="jsk", name="unload")
    async def jsk_unload(self, ctx: commands.Context, *extensions: ExtensionConverter):
        """
        Выгружает заданные имена расширения.

        Сообщает о любых расширениях, которые не разгрузили.
        """

        paginator = WrappedPaginator(prefix='', suffix='')
        icon = "\N{OUTBOX TRAY}"

        for extension in itertools.chain(*extensions):
            try:
                self.bot.unload_extension(extension)
            except Exception as exc:
                traceback_data = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__, 1))

                paginator.add_line(
                    f"{icon}\N{WARNING SIGN} `{extension}`\n```py\n{traceback_data}\n```",
                    empty=True
                )
            else:
                paginator.add_line(f"{icon} `{extension}`", empty=True)

        for page in paginator.pages:
            await ctx.send(page)

    @Feature.Command(parent="jsk", name="shutdown", aliases=["logout"])
    async def jsk_shutdown(self, ctx: commands.Context):
        """
        вырубает этого бот.
        """

        ellipse_character = "\N{BRAILLE PATTERN DOTS-356}" if Flags.USE_BRAILLE_J else "\N{HORIZONTAL ELLIPSIS}"

        await ctx.send(f"Выход из системы...{ellipse_character}")
        await ctx.bot.close()

    @Feature.Command(parent="jsk", name="restart", aliases=["reboot"])
    async def jsk_reboot(self, ctx: commands.Context):
        """
        перезагрузить.
        """

        ellipse_character = "\N{BRAILLE PATTERN DOTS-356}" if Flags.USE_BRAILLE_J else "\N{HORIZONTAL ELLIPSIS}"

        await ctx.send(f"Перезагрузка...{ellipse_character}")
        await ctx.bot.close()
        os.execv(sys.executable, ['python'] + sys.argv)
        await ctx.send("Перезагрузка завершена.")

    @Feature.Command(parent="jsk", name="invite")
    async def jsk_invite(self, ctx: commands.Context, *perms: str):
        """
        Получить URL приглашения для этого бота.

        Если имена разрешений предоставляются, они запрашиваются как часть приглашения.
        """

        scopes = ('bot', 'applications.commands')
        permissions = disnake.Permissions()

        for perm in perms:
            if perm not in dict(permissions):
                raise commands.BadArgument(f"Недопустимое разрешение: {perm}")

            setattr(permissions, perm, True)

        application_info = await self.bot.application_info()

        query = {
            "client_id": application_info.id,
            "scope": "+".join(scopes),
            "permissions": permissions.value
        }

        return await ctx.send(
            f"Ссылка на добавление бота:\n<https://discordapp.com/oauth2/authorize?{urlencode(query, safe='+')}>"
        )

    @Feature.Command(parent="jsk", name="rtt", aliases=["ping"])
    async def jsk_rtt(self, ctx: commands.Context):
        """
        Рассчитывает время запроса к API.
        """

        message = None

        # Мы покажем каждое из этих показаний, а также среднее и стандартное отклонение.
        api_readings = []
        # Мы также записываем показания WebSocket, но мы предоставим только среднее значение.
        websocket_readings = []

        # Мы делаем 6 итераций здесь.
        # Это дает нам 5 видимых показаний, потому что запрос не может включать статистику для себя.
        for _ in range(6):
            # Сначала генерируйте текст
            text = "Расчет времени в пути туда и обратно...\n\n"
            text += "\n".join(f"Чтение {index + 1}: {reading * 1000:.2f}ms" for index, reading in enumerate(api_readings))

            if api_readings:
                average = sum(api_readings) / len(api_readings)

                if len(api_readings) > 1:
                    stddev = math.sqrt(sum(math.pow(reading - average, 2) for reading in api_readings) / (len(api_readings) - 1))
                else:
                    stddev = 0.0

                text += f"\n\nСреднее: {average * 1000:.2f} \N{PLUS-MINUS SIGN} {stddev * 1000:.2f}ms"
            else:
                text += "\n\nПоказаний пока нет."

            if websocket_readings:
                average = sum(websocket_readings) / len(websocket_readings)

                text += f"\nЗадержка вебсокета: {average * 1000:.2f}ms"
            else:
                text += f"\nЗадержка вебсокета: {self.bot.latency * 1000:.2f}ms"

            # Теперь сделайте фактический запрос и чтение
            if message:
                before = time.perf_counter()
                await message.edit(content=text)
                after = time.perf_counter()

                api_readings.append(after - before)
            else:
                before = time.perf_counter()
                message = await ctx.send(content=text)
                after = time.perf_counter()

                api_readings.append(after - before)

            # Игнорируйте задержки WebSocket, которые 0 или негативны, потому что они обычно означают, что у нас плохое сердцебиение
            if self.bot.latency > 0.0:
                websocket_readings.append(self.bot.latency)

    @Feature.Command(parent="jsk", name="help", aliases=["commands"])
    async def jsk_help(self, ctx: disnake.ApplicationCommandInteraction):
        """
        Список доступных команд jishaku
        """

        commands_info = {
            "cancel": "Отменяет задачу с указанным индексом.",
            "cat": "Читает файл, используя подсветку синтаксиса.",
            "curl": "Скачивает и отображает текстовый файл из интернета.",
            "debug": "Запускает команду, измеряя время выполнения.",
            "dis": "Дизассемблирует код Python в байт-код.",
            "git": "Сокращение для 'jsk sh git'. Вызывает системную оболочку.",
            "hide": "Скрывает Jishaku из команды help.",
            "invite": "Получает URL-адрес приглашения для этого бота.",
            "load": "Загружает или перезагружает указанные имена расширений.",
            "override": "Запускает команду от имени другого пользователя, канала или потока, с до... ",
            "permtrace": "Вычисляет источник предоставленных или отклоненных разрешений.",
            "pip": "Сокращение для 'jsk sh pip'. Вызывает системную оболочку.",
            "py": "Прямая оценка кода Python.",
            "py_inspect": "Оценка кода Python с информацией о проверке.",
            "repeat": "Запускает команду несколько раз подряд.",
            "retain": "Включает или отключает сохранение переменных для REPL.",
            "rtt": "Вычисляет время двусторонней передачи данных до API.",
            "shell": "Выполняет команды в системной оболочке.",
            "show": "Показывает Jishaku в команде help.",
            "shutdown": "Выводит этого бота из системы.",
            "source": "Отображает исходный код для команды.",
            "tasks": "Показывает запущенные задачи jishaku.",
            "unload": "Отключает указанные имена расширений.",
            "voice": "Команды, связанные с голосом.",
        }

        commands_list = list(commands_info.items())
        mid_index = len(commands_list) // 2
        page1_commands = commands_list[:mid_index]
        page2_commands = commands_list[mid_index:]

        is_embedded = os.environ.get('JISHAKU_EMBEDDED_JSK', 'false').lower() == 'true'
        
        current_page = 1  

        if is_embedded:
            embed = disnake.Embed(title="Команды | 1 Страница", description="Список доступных команд")
            for command, description in page1_commands:
                embed.add_field(name=command, value=description, inline=False)
            message = await ctx.send(embed=embed)
        else:
            text_response = "Команды | 1 Страница:\n" + "\n".join(f"**{command}**: {description}" for command, description in page1_commands)
            message = await ctx.send(text_response)

        button_page1 = disnake.ui.Button(label="<—", style=disnake.ButtonStyle.secondary, disabled=True)
        button_page2 = disnake.ui.Button(label="—>", style=disnake.ButtonStyle.secondary)

        view = disnake.ui.View()
        view.add_item(button_page1)
        view.add_item(button_page2)

        async def update_message(page):
            nonlocal current_page
            current_page = page

            if current_page == 1:
                button_page1.disabled = True
                button_page2.disabled = False
                commands_to_display = page1_commands
            else:
                button_page1.disabled = False
                button_page2.disabled = True
                commands_to_display = page2_commands

            if is_embedded:
                embed.title = f"Команды | {current_page} Страница" 
                embed.clear_fields()
                for command, description in commands_to_display:
                    embed.add_field(name=command, value=description, inline=False)
                await message.edit(embed=embed, view=view)
            else:
                text_response = f"Команды | {current_page} Страница:\n" + "\n".join(f"**{command}**: {description}" for command, description in commands_to_display)
                await message.edit(content=text_response, view=view)

        async def button_page1_callback(interaction: disnake.Interaction):
            if interaction.user != ctx.author:
                await interaction.response.send_message("Это не ваша кнопка!", ephemeral=True)
                return
            await update_message(1)
            await interaction.response.defer()

        async def button_page2_callback(interaction: disnake.Interaction):
            if interaction.user != ctx.author:
                await interaction.response.send_message("Это не ваша кнопка!", ephemeral=True)
                return
            await update_message(2)
            await interaction.response.defer()

        button_page1.callback = button_page1_callback
        button_page2.callback = button_page2_callback

        await message.edit(view=view)
